__version__ = "0.999"
__author__ = "Paolo Morettin"

from subprocess import call

from wmipa.integration.command_line_integrator import CommandLineIntegrator
from wmipa.integration.polytope import Polynomial, Polytope
from wmipa.wmiexception import WMIRuntimeException


class LatteIntegrator(CommandLineIntegrator):
    """This class handles the integration of polynomial functions over (convex) polytopes.

    It inherits from the abstract class CacheIntegrator.

    LattE Integrale is required.

    Attributes:
        algorithm (str): The algorithm to use when computing the integrals.
        n_threads (int): The number of threads to use.

    """

    ALG_TRIANGULATE = "--triangulate"
    ALG_CONE_DECOMPOSE = "--cone-decompose"
    DEF_ALGORITHM = ALG_CONE_DECOMPOSE

    ALGORITHMS = [ALG_TRIANGULATE, ALG_CONE_DECOMPOSE]

    def __init__(self, algorithm=DEF_ALGORITHM, **options):
        """Default constructor.

        It calls the init method of the parent.

        Args:
            algorithm (str): The algorithm to use when computing the integrals.
            options: @see CommandLineIntegrator.__init__

        """
        super().__init__(**options)
        self.algorithm = algorithm

        # check that algorithm exists
        if self.algorithm not in self.ALGORITHMS:
            err = "{}, choose one from: {}".format(
                self.algorithm, ", ".join(self.ALGORITHMS)
            )
            raise WMIRuntimeException(WMIRuntimeException.INVALID_MODE, err)

    @classmethod
    def _make_problem(cls, weight, bounds, aliases):
        """Makes the problem to be solved by LattE.
        Args:
            weight (FNode): The weight function.
            bounds (list): The polytope.
            aliases (dict): The aliases of the variables.
        Returns:
            integrand (Polynomial): The integrand.
            polytope (Polytope): The polytope.
        """
        integrand = Polynomial(weight, aliases)
        polytope = Polytope(bounds, aliases)

        return integrand, polytope

    def _write_integrand_file(self, integrand, variables, path):
        """Writes the integrand to the given file.

        Args:
            integrand (Polynomial): The integrand.
            variables (list): The list of variables.
            path (str): The path where to write the file.

        """
        # Create the string representation of the integrand (LattE format)
        monomials_repr = []
        for monomial in integrand.monomials:
            monomial_repr = "[" + str(monomial.coefficient) + ",["
            exponents = []
            for var in variables:
                if var in monomial.exponents:
                    exponents.append(str(monomial.exponents[var]))
                else:
                    exponents.append("0")
            monomial_repr += ",".join(exponents) + "]]"
            monomials_repr.append(monomial_repr)
        latte_repr = "[" + ",".join(monomials_repr) + "]"

        # Write the string on the file
        with open(path, "w") as f:
            f.write(latte_repr)

    def _call_integrator(self, polynomial_file, polytope_file, output_file):
        """Calls LattE executable to calculate the integrand of the problem
            represented by the given files.

        Args:
            polynomial_file (str): The path where to find the file with the
                representation of the polynomial.
            polytope_file (str): The path where to find the file with representation
                of the polytope.
            output_file (str): The file where to write the result of the computation.

        """
        cmd = [
            "integrate",
            "--valuation=integrate",
            self.algorithm,
            "--monomials=" + polynomial_file,
            polytope_file,
        ]

        with open(output_file, "w") as f:
            if self.stub_integrate:
                f.write("")
            else:
                return_value = call(cmd, stdout=f, stderr=f)
                if return_value != 0:
                    # print(open(output_file).read())
                    """
                    if return_value != 0:
                        msg = "LattE returned with status {}"
                        # LattE returns an exit status != 0 if the polytope is empty.
                        # In the general case this may happen, raising an exception
                        # is not a good idea.
                    """



"""This module leverages sympy to transform a pysmt polynomial in canonical form.

"""

__version__ = "0.99"
__author__ = "Paolo Morettin"

from pysmt.shortcuts import Plus, Pow, Real, Symbol, Times, serialize
from pysmt.typing import REAL
from sympy import SympifyError, expand, sympify

from wmipa.wmiexception import WMIParsingException


def pysmt2sympy(expression):
    """Converts a pysmt formula representing a polynomial into a string.
        The string can then be read and modified by sympy.

    Args:
        expression (FNode): The pysmt formula to convert.

    Returns:
        str: The string representing the formula.

    Raises:
        WMIParsingException: If the method fails to parse the formula.

    """
    # write reals with no fractional part as integers to help sympy group monomials
    serialize_formula = serialize(expression).replace(".0", "")
    try:
        sympy_formula = sympify(serialize_formula)
    except SympifyError:
        raise WMIParsingException(
            WMIParsingException.CANNOT_CONVERT_PYSMT_FORMULA_TO_SYMPY, expression
        )
    return sympy_formula


def sympy2pysmt(expression):
    """Converts a sympy formula representing a polynomial into a pysmt formula.

    Args:
        expression: The sympy formula to convert.

    Returns:
        FNode: The pysmt formula.

    Raises:
        WMIParsingException: If the method fails to parse the formula.

    """
    if expression.is_Add:
        return Plus(map(sympy2pysmt, expression.args))
    elif expression.is_Mul:
        return Times(map(sympy2pysmt, expression.args))
    elif expression.is_Pow:
        base, exp = expression.args
        return Pow(sympy2pysmt(base), sympy2pysmt(exp))
    elif expression.is_Symbol:
        return Symbol(str(expression), REAL)
    elif expression.is_Number:
        return Real(float(expression))
    else:
        raise WMIParsingException(
            WMIParsingException.CANNOT_CONVERT_SYMPY_FORMULA_TO_PYSMT, expression
        )


def get_canonical_form(expression):
    """Given a pysmt formula representing a polynomial, rewrites it in canonical form.

    Args:
        expression (FNode): The pysmt formula to rewrite.

    Returns:
        (FNode): The pysmt formula in canonical form.

    Raises:
        WMIParsingException: If the method fails to parse the formula.

    """
    sympy_formula = pysmt2sympy(expression)
    sympy_expand = expand(sympy_formula)
    canonical = sympy2pysmt(sympy_expand)
    return canonical



# check whether MathSAT is installed or not
try:
    import mathsat
except ImportError:
    IMPORT_ERR_MSG = """Couldn't import mathsat.
MathSAT needs to be manually installed via:
    pysmt-install --msat"""

from .log import logger
from .wmi import WMI




# variable definitions

def RealVar(name):
    return "random {} : Real;".format(name)

def BooleanVar(name):
    return "random {} : Boolean;".format(name)


# constant values
def Real(x):
    return str(float(x))

def Boolean(x):
    return "true" if x else "false"

# if-then-elses
def Ite(cond, then, else_):
    assert(isinstance(cond, str))
    assert(isinstance(then, str))
    assert(isinstance(else_, str))    
    return "(if {}\nthen {}\nelse {})".format(cond, then, else_)


# generic n-ary operator
def bOp(a, b, op):
    assert(isinstance(a, str))
    assert(isinstance(b, str))
    assert(isinstance(op, str))
    op = " {} ".format(op.strip())
    return "({} {} {})".format(a, op, b)

# generic n-ary operator
def nOp(vals, op):
    assert(all(map(lambda x : isinstance(x, str), vals)))
    assert(isinstance(op, str))    
    op = " {} ".format(op.strip())
    return "({})".format(op.join(vals))

# algebraic operators
            
def Plus(vals):
    return nOp(vals, "+")
    
def Minus(vals):
    return nOp(vals, "-")
    
def Times(vals):
    return nOp(vals, "*")

def Pow(a, b):
    return bOp(a, b, "^")


# boolean operators
    
def And(vals):
    return nOp(vals, "and")
        
def Or(vals):
    return nOp(vals, "or")    

def Not(a):
    return "(not {})".format(a)

def Iff(a, b):
    return bOp(a, b, "<=>")
    
def Implies(a, b):
    return bOp(a, b, "=>")

# theory relations

def LE(a, b):
    return bOp(a, b, "<=")

def LT(a, b):
    return bOp(a, b, "<")

def GE(a, b):
    return bOp(a, b, ">=")

def GT(a, b):
    return bOp(a, b, ">")

def Equals(a, b):
    return bOp(a, b, "=")


def AtMostOne(args):
    excl = []
    for i in range(len(args) - 1):
        for j in range(i+1, len(args)):
            not_both = Not(And([args[i], args[j]]))
            excl.append(not_both)
    return And(excl)
    
def ExactlyOne(terms):    
    return And([Or(terms), AtMostOne(terms)])



from fractions import Fraction

from pysmt.shortcuts import Real, Times, Pow, Symbol, Plus
from pysmt.typing import REAL

from wmipa.integration.integrand import Integrand
from wmipa.integration.sympy2pysmt import get_canonical_form
from wmipa.utils import is_pow, apply_aliases
from wmipa.wmiexception import WMIParsingException


class Monomial:
    """Intermediate representation of a monomial.

    c * x_1^e_1 * ... * x_k^e_k

    Attributes:
        coefficient (Fraction): The coefficient of the monomial.
        exponents (dict {FNode : Fraction}): The dictionary containing the
            exponents {x_i : e_i}

    """

    def __init__(self, expression):
        """Default constructor.

        Takes as input a pysmt formula representing a monomial.

        Args:
            expression (FNode): The pysmt formula representing a monomial.

        Raises:
            WMIParsingException: If the expression is not a monomial.

        """
        if not (
                expression.is_times()
                or expression.is_real_constant()
                or expression.is_symbol()
                or is_pow(expression)
        ):
            raise WMIParsingException(WMIParsingException.NOT_A_MONOMIAL, expression)

        self.coefficient, self.exponents = self._parse_sub(expression)

    def __str__(self):
        """The str method.

        Returns:
            str: The string representation of the monomial (e.g: (-6 * x^3 * y^2) ).

        """
        powers = ["{}^{}".format(k, e) for k, e in self.exponents.items()]
        if len(self.exponents) == 0:
            return "(" + str(self.coefficient) + ")"
        else:
            return "(" + str(self.coefficient) + " * " + " * ".join(powers) + ")"

    def degree(self):
        """Calculates the degree of the monomial.

        This method evaluates the degree by summing up all the values of the exponents.

        Returns:
            int: The degree of the monomial.

        """
        return sum(self.exponents.values())

    def multiply_by_monomial(self, monomial):
        """Multiplies this instance with another monomial.

        Args:
            monomial (FNode): The other monomial.

        """
        assert isinstance(
            monomial, Monomial
        ), "Argument should be an instance of Monomial"

        # multiply the coefficients
        self.coefficient *= monomial.coefficient

        # update all the exponents
        for name, exp in monomial.exponents.items():
            if name not in self.exponents:
                self.exponents[name] = Fraction(0)
            self.exponents[name] += exp

    def negate(self):
        """Negates the monomial by changing the sign of the coefficient."""
        self.coefficient *= -1

    def _parse_sub(self, expression):
        """Parse an expression representing a monomial.

        Args:
            expression (FNode): The pysmt formula representing a monomial.

        Returns:
            Fraction: The coefficient of the expression.
            (dict {FNode : Fraction}): The dictionary containing the exponents.

        Raises:
            WMIParsingException: If the expression is not a monomial.

        """

        if expression.is_real_constant():
            return expression.constant_value(), {}
        elif expression.is_symbol():
            return Fraction(1), {expression.symbol_name(): Fraction(1)}
        elif is_pow(expression):
            base, exp = expression.args()

            # Check that the exponent is constant
            if not exp.is_real_constant():
                raise WMIParsingException(
                    WMIParsingException.NOT_A_MONOMIAL, expression
                )
            exp_value = exp.constant_value()

            # Parse the base
            base_value, base_var = self._parse_sub(base)

            ret_value = pow(base_value, exp_value)
            for x in base_var.keys():
                base_var[x] *= exp_value

            return ret_value, base_var
        elif expression.is_times():
            ret_value = Fraction(1)
            ret_var = {}

            # Parse all the terms of the multiplication
            for sub in expression.args():
                sub_value, sub_var = self._parse_sub(sub)
                ret_value *= sub_value

                for x, v in sub_var.items():
                    if x not in ret_var:
                        ret_var[x] = 0
                    ret_var[x] += v
            return ret_value, ret_var
        else:
            raise WMIParsingException(WMIParsingException.NOT_A_MONOMIAL, expression)

    def to_pysmt(self):
        if not self.exponents:
            return Real(self.coefficient)
        return Times(
            Real(self.coefficient),
            *map(
                lambda i: Pow(Symbol(i[0], REAL), Real(i[1])),
                self.exponents.items(),
            ),
        )


class Polynomial(Integrand):
    """Intermediate representation of a polynomial.

    Attributes:
        monomials (Monomial): The list of Monomial instances.
        variables (set(FNode)): The set of all the variables of the polynomial.

    """

    def __init__(self, expression, aliases=None):
        """Default constructor.

        Takes as input a pysmt formula representing a polynomial and a dict of
        aliases to be substituted before the parsing.

        Args:
            expression (FNode): The pysmt formula representing the polynomial.
            aliases (dict {FNode : FNode}): The dict containing the aliases definitions.

        """
        super().__init__()
        # Perform aliases substitution and put in canonical form
        if aliases is None:
            aliases = {}
        canonical = Polynomial._preprocess_formula(expression, aliases)

        # Process every monomial individually
        self.monomials = []
        if canonical.is_plus():
            for term in canonical.args():
                self._add_monomial(Monomial(term))
        else:
            self._add_monomial(Monomial(canonical))

    def __str__(self):
        """The str method.

        Returns:
            str: The string representation of the polynomial
                (e.g: (3 * x^5) + (-2 * y^3)).

        """
        return " + ".join(map(str, self.monomials))

    def degree(self):
        """Calculates the degree of the polynomial.

        This method evaluates the degree by taking the maximum degree of its monomials.

        Returns:
            int: The degree of the polynomial.

        """
        if len(self.monomials) > 0:
            return max([m.degree() for m in self.monomials])
        else:
            return 0

    def negate(self):
        """Negates the polynomial by negating all its monomials."""
        for monomial in self.monomials:
            monomial.negate()

    def _add_monomial(self, monomial):
        """Add a monomial to the polynomial.

        Args:
            monomial (Monomial): The monomial to be added.

        """
        assert isinstance(
            monomial, Monomial
        ), "Argument should be an instance of Monomial"
        if monomial.coefficient != 0:
            self.monomials.append(monomial)
            self.variables = self.variables.union(set(monomial.exponents.keys()))

    @staticmethod
    def _preprocess_formula(expression, aliases):
        """Preprocesses a pysmt polynomial by substituting the aliases and
        rewriting it in canonical form.

        Args:
            expression (FNode): The pysmt formula representing a polynomial.
            aliases (dict {FNode : FNode}): The dict containing the aliases definitions.

        Returns:
            FNode: The pysmt formula in canonical form with all the aliases substituted.

        Raises:
            WMIParsingException: If in the aliases there is a cyclic assignment
                (e.g: x = y, y = x).
        """
        # Apply alias substitutions in the correct order
        expression = apply_aliases(expression, aliases)

        return get_canonical_form(expression)

    def to_pysmt(self):
        if not self.monomials:
            return Real(0)
        return Plus(*map(lambda x: x.to_pysmt(), self.monomials))



from abc import abstractmethod
from os import chdir, getcwd
from tempfile import TemporaryDirectory

from wmipa.integration.cache_integrator import CacheIntegrator
from wmipa.wmiexception import WMIRuntimeException

class WMICommandLineIntegratorException(WMIRuntimeException):
    """
    Exception raised when the command line integrator fails.
    """

    MEMORY_LIMIT = 0

    messages = {
        MEMORY_LIMIT: "Memory limit exceeded",
    }

    def __init__(self, code, value=None):
        """
        Default constructor.

        It calls the init method of the parent.

        Args:
            code (int): The code of the exception.
            value: Additional info about the value that raised the exception (default: None).

        """
        super().__init__(code, value)

class CommandLineIntegrator(CacheIntegrator):
    DEF_ALGORITHM = None
    ALGORITHMS = []
    # Template name for the temporary folder
    FOLDER_TEMPLATE = "tmp_{}"

    # Temporary files
    POLYTOPE_TEMPLATE = "polytope.hrep"
    INTEGRAND_TEMPLATE = "integrand.txt"
    OUTPUT_TEMPLATE = "output.txt"

    def __init__(self, **options):
        super().__init__(**options)

    def _integrate_problem(self, integrand, polytope):
        """Generates the input files and calls integrator executable
            to calculate the integral. Then, reads back the result and returns it
            as a float.

        Args:
            integrand (Integrand): The integrand of the integration.
            polytope (Polytope): The polytope of the integration.

        Returns:
            real: The integration result.

        """
        # Create a temporary folder containing the input and output files
        # possibly removing an older one
        with TemporaryDirectory(dir=".") as folder:
            integrand_file = self.INTEGRAND_TEMPLATE
            polytope_file = self.POLYTOPE_TEMPLATE
            output_file = self.OUTPUT_TEMPLATE

            # Change the CWD
            original_cwd = getcwd()
            chdir(folder)

            # Variable ordering is relevant in LattE files
            variables = sorted(integrand.variables.union(polytope.variables))

            # Write integrand and polytope to file
            self._write_integrand_file(integrand, variables, integrand_file)
            self._write_polytope_file(polytope, variables, polytope_file)
            # key = tuple([polytope_key, cond_assignments])

            # if cache:
            #     value = self.hashTable.get(key)
            #     if value is not None:
            #         chdir(original_cwd)
            #         return value, 1

            # Integrate and dump the result on file
            self._call_integrator(integrand_file, polytope_file, output_file)

            # Read back the result and return to the original CWD
            result = self._read_output_file(output_file)
            chdir(original_cwd)

            return result

    def _read_output_file(self, path):
        """Reads the output file generated by the computation of the integrator.

        Args:
            path (str): The path of the file to read.

        Returns:
            real: The result of the computation written in the file.

        """
        res = None

        with open(path, "r") as f:
            lines = f.readlines()
            for line in lines:
                # Result in the "Answer" line may be written in fraction form
                if "Decimal" in line:
                    # print("Res: {}".format(line))
                    return float(line.partition(": ")[-1].strip())

            # error (possibly interrupted due to memory limit)
            raise WMICommandLineIntegratorException(WMICommandLineIntegratorException.MEMORY_LIMIT)

        return res

    @abstractmethod
    def _write_integrand_file(self, integrand, variables, path):
        """Writes the integrand into a file from where the integrator will read.

        Args:
            integrand (Integrand): The integrand of the integration.
            variables (list): The sorted list of all the variables involved in the
                integration.
            path (str): The path of the file to write.
        """
        pass

    def _write_polytope_file(self, polytope, variables, path):
        """Writes the polytope into a file from where the integrator will read.

        Args:
            polytope (Polytope): The polytope of the integration.
            variables (list): The sorted list of all the variables involved in the
                integration.
            path (str): The path of the file to write.
        """
        # Create the string representation of the polytope (LattE format)
        n_ineq = str(len(polytope.bounds))
        n_vars = str(len(variables) + 1)
        latte_repr = "{} {}\n".format(n_ineq, n_vars)
        for _, bound in enumerate(polytope.bounds):
            latte_repr += str(bound.constant) + " "
            for var in variables:
                if var in bound.coefficients:
                    latte_repr += str(-bound.coefficients[var]) + " "
                else:
                    latte_repr += "0 "
            latte_repr += "\n"

        # Write the string on the file
        with open(path, "w") as f:
            f.write(latte_repr)

    @abstractmethod
    def _call_integrator(self, polynomial_file, polytope_file, output_file):
        pass


"""This module leverages PRAiSE to implement an inference engine that
allows queries of the form P(Q(x,A)|E(x,A)) over an hybrid model
expressed by a weight formula w(x,A) and support Chi(x,A). Q, E, w and
Chi are SMT-LRA formulas.

"""
__version__ = '0.999'
__author__ = 'Paolo Morettin'

from os.path import abspath, dirname, join
from subprocess import Popen, PIPE
from time import sleep
from fractions import Fraction
from pysmt.typing import REAL, BOOL
import sys
from tempfile import TemporaryDirectory
import wmipa.praiselang as pl
from wmipa.wmiexception import WMIRuntimeException, WMIParsingException
from wmipa.utils import is_pow


class PRAiSEInference:

    DEF_TIMEOUT = 10000
    POLL_DELTA = 1

    JAR_NAME = "praise.jar"
    JAR_PATH = dirname(abspath(__file__)) + "/thirdparties/" + JAR_NAME

    def __init__(self, support, weights, timeout=None):
        self.timeout = timeout or PRAiSEInference.DEF_TIMEOUT
        self.convert(support, weights)

    def convert(self, formula, weights):
        self.variables = {}
        converted_formula = PRAiSEInference.convert_formula(formula, self.variables)
        converted_weights = PRAiSEInference.convert_formula(weights, self.variables)
        definitions = PRAiSEInference._get_definitions(self.variables)
        self.model = "\n\n".join([definitions,
                            converted_formula + ";",
                            converted_weights + ";"])
    
    def compute_normalized_probability(self, query, evidence=None):

        model_evidence = self.model
        query_variables = dict(self.variables)
        if evidence:
            evidence = PRAiSEInference.convert_formula(evidence, query_variables)
            model_evidence += "\n\n" + evidence +";\n"

        query = PRAiSEInference.convert_formula(query, query_variables)
        new_variables = {k : v for k,v in query_variables.items() if k not in self.variables.keys()}
        new_vars_definition = PRAiSEInference._get_definitions(new_variables)
        model_evidence += "\n\n" + new_vars_definition

        with TemporaryDirectory(dir=".") as folder:

            temp_path = join(abspath(folder), "model.txt")
            with open(temp_path, "w") as f:
                f.write(model_evidence)

            tout = int(self.timeout / PRAiSEInference.POLL_DELTA)
            task = Popen(["java", "-jar", PRAiSEInference.JAR_PATH, temp_path,
            "--query", query], stdout=PIPE, stderr=PIPE)

            # polling every POLL_DELTA seconds
            while task.poll() is None and tout > 0:
                sleep(PRAiSEInference.POLL_DELTA)
                tout -= PRAiSEInference.POLL_DELTA

            if task.poll() != None:
                out, err = task.communicate()

                err = err.decode(sys.stdout.encoding)
                out = out.decode(sys.stdout.encoding)

                if len(err) > 0:
                    msg = "PRAiSE error: {}"
                    raise WMIRuntimeException(msg.format(err))

                return PRAiSEInference._parse_result(out)
        
            else:
                task.terminate()
                raise Exception("PRAiSE timed out")
        
    @staticmethod
    def _get_definitions(variables):
        definitions = []        
        for var_name, var_type in variables.items():
            if var_type == BOOL:
                definitions.append(pl.BooleanVar(var_name))
            elif var_type == REAL:
                definitions.append(pl.RealVar(var_name))
            else:
                raise WMIParsingError("Unknown type", None)
        return "\n".join(definitions)

    @staticmethod
    def _parse_result(praise_out):
        for line in praise_out.split("\n"):
            line = line.strip()
            if line.startswith("Result"):
                result_str = line.partition(": ")[-1].strip()
                try :
                    numerator, denominator = result_str.split("/")
                    numerator = int(float(numerator))
                    denominator = int(float(denominator))
                    return float(Fraction(numerator, denominator))
                except ValueError:
                    return float(result_str)
    
    @staticmethod
    def convert_formula(formula, variables):

        # 0-ariety expressions
        if formula.is_constant():
            return str(formula.constant_value()).lower()
        
        elif formula.is_symbol():
            # whenever a variable is found, return its name and optionally add
            # an entry to the variables dictionary with the corresponding type
            var_name = formula.symbol_name()
            var_type = formula.get_type()
            if var_name not in variables:
                variables[var_name] = var_type
            return var_name
        
        else:
            # recursively convert subformulas
            args = []
            for arg in formula.args():
                args.append(PRAiSEInference.convert_formula(arg, variables))

            # boolean operators
            if formula.is_not():
                assert(len(args) == 1)
                return pl.Not(args[0])
            
            elif formula.is_and():
                return pl.And(args)
            
            elif formula.is_or():
                return pl.Or(args)
            
            elif formula.is_implies():
                assert(len(args) == 2)
                return pl.Implies(args[0], args[1])
            
            elif formula.is_iff():
                assert(len(args) == 2)
                return pl.Iff(args[0], args[1])

            # theory relations
            elif formula.is_le():
                assert(len(args) == 2)
                return pl.LE(args[0], args[1])
            
            elif formula.is_lt():
                assert(len(args) == 2)
                return pl.LT(args[0], args[1])
            
            elif formula.is_equals():
                assert(len(args) == 2)
                return pl.Equals(args[0], args[1])

            # algebraic operators
            if formula.is_plus():
                return pl.Plus(args)
            
            elif formula.is_minus():
                return pl.Minus(args)
            
            elif formula.is_times():
                return pl.Times(args)
            
            elif is_pow(formula):
                assert(len(args) == 2)
                return pl.Pow(args[0], args[1])

            # if-then-else
            elif formula.is_ite():
                assert(len(args) == 3)
                return pl.Ite(args[0], args[1], args[2])
            
            else:
                raise WMIParsingError("Unhandled formula format", formula)


if __name__ == "__main__":
    from pysmt.shortcuts import Symbol, Ite, And, LE, LT, Real, Times, serialize
    from pysmt.typing import REAL

    def compute_print(method, query, evidence):
        print("query: ", serialize(query))
        print("evidence: ", serialize(evidence) if evidence else "-")
        prob = method.compute_normalized_probability(query, evidence)
        print("normalized: ", prob)
        print("--------------------------------------------------")

    x = Symbol("x", REAL)
    A = Symbol("A")
    support = And(LE(Real(-1), x), LE(x, Real(1)))
    weights = Ite(LT(Real(0), x),
                  Ite(A, Times(Real(2), x), x),
                  Ite(A, Times(Real(-2), x), Times(Real(-1),x)))

    praise = PRAiSEInference(support, weights)
    print("support: ", serialize(support))
    print("weights: ", serialize(weights))
    print("==================================================")

    suite = [(A, None),
             (And(A, LE(Real(0), x)), None),
             (LE(Real(0), x), A)]

    for query, evidence in suite:
        compute_print(praise, query, evidence)



"""This module implements the exceptions used throughout the code.

"""

version = "0.99"
author = "Paolo Morettin"


class WMIException(Exception):
    """This class represents the general exception used in the WMI module.
        Every other exception will inherit from this one.

    Attributes:
        message (str): Human readable string describing the exception.

    """

    def __init__(self, message):
        """Default constructor.

        It assigns the message to the attributes.

        Args:
            message (str): Human readable string describing the exception.

        """
        self.message = message

    def __str__(self):
        """The str method.

        It uses the 'str' method to display the message of the exception.

        Returns:
            str: The string representation of the message.
        """
        return str(self.message)


class WMIParsingException(WMIException):
    """This exception handles all the cases where the code fails to parse a given expression of formula.

    Attributes:
        code (int): The code of the exception.
        value: Additional info about the value that raised the exception (default: None).

    """

    """During the parsing of the formula, it was encountered an equality that does not respect
        the specific formatting of an equality. One of the two arguments must return True to the
        pysmt method '.is_symbol()' and the other one must return True to the expression
        '.get_type() == REAL'.
        An example of correct equality is: 'x = Plus(y, Real(3))' where
        x = Symbol('X', REAL) and y = Symbol('Y', REAL).

    """
    MALFORMED_ALIAS_EXPRESSION = 0

    """During the parsing of the formula, it was encountered a node that cannot exists inside
        the formula representing the weight function. The weight function admits only these internal nodes:
        PLUS, MINUS, TIMES, DIV, POW, ITE. Also as a leaf node there must be a Real or a Symbol with
        type equal to Real.

    """
    INVALID_WEIGHT_FUNCTION = 1

    """During the parsing of the formula, it was encountered an expression that couldn't be converted to pysmt.
        The expression must be of one of this types: PLUS, MINUS, TIMES, DIV, POW, SYMBOL, REAL.

    """
    CANNOT_CONVERT_SYMPY_FORMULA_TO_PYSMT = 2

    """During the parsing of the formula, it was encountered an expression that couldn't be converted to sympy.
        The expression must be of one of this types: PLUS, MINUS, TIMES, DIV, POW, SYMBOL, REAL.

    """
    CANNOT_CONVERT_PYSMT_FORMULA_TO_SYMPY = 3

    """During the parsing of the formula, it was encountered an expression that is not a monomial.

    """
    NOT_A_MONOMIAL = 4

    """During the parsing of the formula, it was encountered two or more aliases (equalities) that
        lead to a cyclic assignment (e.g: y = x, x = y).

    """
    CYCLIC_ASSIGNMENT_IN_ALIASES = 5

    """During the parsing of the formula, it was encountered an expression that does not represent an inequality.

    """
    NOT_AN_INEQUALITY = 6

    """During the parsing of the formula, it was encountered an expression that represents a polynomial
        with a degree greater than one. This is not supported in this module because it handles only
        linear inequalities over reals.

    """
    POLYNOMIAL_WITH_DEGREE_GREATER_THAN_ONE = 7

    """During the parsing of the formula, it was encountered a same alias with two or more assignments
        (e.g: z = x+3, z = y-2). This module does not support this kind of operations, for every alias
        there must be at most one assignment.

    """
    MULTIPLE_ASSIGNMENT_SAME_ALIAS = 8

    messages = {
        MALFORMED_ALIAS_EXPRESSION: "Malformed alias expression",
        INVALID_WEIGHT_FUNCTION: "Invalid weight function",
        CANNOT_CONVERT_SYMPY_FORMULA_TO_PYSMT: "Cannot convert the sympy formula to a pysmt formula",
        CANNOT_CONVERT_PYSMT_FORMULA_TO_SYMPY: "Cannot convert the pysmt formula to a sympy formula",
        NOT_A_MONOMIAL: "Not a monomial",
        CYCLIC_ASSIGNMENT_IN_ALIASES: "Cyclic assignment in the aliases",
        NOT_AN_INEQUALITY: "Not an inequality",
        POLYNOMIAL_WITH_DEGREE_GREATER_THAN_ONE: "Polynomial with degree greater than one",
        MULTIPLE_ASSIGNMENT_SAME_ALIAS: "Multiple assignments to the same alias",
    }

    def __init__(self, code, value=None):
        """Default constructor.

        It first calls the init method of the parent and then assigns the expression to the attributes.

        Args:
            code (int): The code of the exception.
            value (optional): Additional info about the value that raised the exception (default: None).

        """
        self.code = code
        self.value = value
        if value:
            message = "{}: {}".format(self.messages[code], self.value)
        else:
            message = self.messages[code]
        super().__init__(message)


class WMIRuntimeException(WMIException):
    """This exception handles all the cases where the code fails because of wrong parameters or settings.

    Attributes:
        code (int): The code of the exception.
        value: Additional info about the value that raised the exception (default: None).
    """

    """The domain of integration of the numerical variables should be all the variables in the formula.
        The domain of integration of the Boolean variables should be a superset of all the boolean
        variables in the formula.
    """
    DOMAIN_OF_INTEGRATION_MISMATCH = 0

    """The mode selected is not correct. Check which are the possibilities (it should be displayed
        in the error message).

    """
    INVALID_MODE = 1

    OTHER_ERROR = 2

    messages = {
        DOMAIN_OF_INTEGRATION_MISMATCH: "Domain of integration mismatch",
        INVALID_MODE: "Invalid mode",
        OTHER_ERROR: "RuntimeException",
    }

    def __init__(self, code, value=None):
        """Default constructor.

        It calls the init method of the parent.

        Args:
            code (int): The code of the exception.
            value: Additional info about the value that raised the exception (default: None).

        """
        self.code = code
        self.value = value
        if value:
            message = "{}: {}".format(self.messages[code], self.value)
        else:
            message = self.messages[code]
        super().__init__(message)


class WMIParsingFileException(WMIException):
    """This exception handles all the cases where the code fails parsing the input file.

    Attributes:
        code (int): The code of the exception.
        value: Additional info about the value that raised the exception (default: None).
    """

    """During the parsing of the file it was encountered an operation that is not supported by
        WMI module.

    """
    OPERATION_NOT_SUPPORTED = 0

    """During the parsing of the file it was encountered a type error. More info in the error message
        itself.

    """
    TYPE_ERROR = 1

    """During the parsing of the file it was encountered a double type declaration, that is the same
        variable was declared twice in the file (e.g: y=5; y=10).

    """
    DOUBLE_DECLARATION = 2

    """During the parsing of the file it was encountered a type that is not supported by WMI.

    """
    TYPE_NOT_SUPPORTED = 3

    """During the parsing of the file it was encountered a variable that was not declared in the model.
        More info in the error message itself.

    """
    VARIABLE_NOT_DECLARED = 4

    """During the parsing of the file it was encountered a variable that was not initialized.
        More info in the error message itself.

    """
    VARIABLE_NOT_INITIALIZED = 5

    """During the parsing of the file it was encountered a type that is not supported by WMI.
        More info in the error message itself.

    """
    DOUBLE_WEIGHT_DECLARATION = 6

    """During the parsing of the file it was encountered a query declaration in a model file.
        This type of declaration can only be declared in the query or complete file.

    """
    QUERY_IN_MODEL = 7

    """During the parsing of the file it was encountered a syntax error (this could be an error
        in the format itself or something that is not recognized by WMI.
        More info in the error message itself.

    """
    SYNTAX_ERROR = 8

    messages = {
        OPERATION_NOT_SUPPORTED: "Operation not supported",
        TYPE_ERROR: "Type error",
        DOUBLE_DECLARATION: "Variable already defined",
        TYPE_NOT_SUPPORTED: "Type not supported",
        VARIABLE_NOT_DECLARED: "Variable not declared",
        VARIABLE_NOT_INITIALIZED: "Variable not initialized",
        DOUBLE_WEIGHT_DECLARATION: "Double weight declaration",
        QUERY_IN_MODEL: "Query declaration in model file",
        SYNTAX_ERROR: "Syntax error",
    }

    def __init__(self, code, value=None):
        """Default constructor.

        It calls the init method of the parent.

        Args:
            code (int): The code of the exception.
            value: Additional info about the value that raised the exception (default: None).

        """
        self.code = code
        self.value = value
        if value:
            message = "{}: {}".format(self.messages[code], self.value)
        else:
            message = self.messages[code]
        super().__init__(message)



from abc import ABC, abstractmethod
from functools import reduce

from pysmt.rewritings import CNFizer, nnf
from pysmt.shortcuts import (
    And,
    Bool,
    Equals,
    FreshSymbol,
    FunctionType,
    Not,
    Or,
    Real,
    Times,
    get_env, )
from pysmt.simplifier import Simplifier
from pysmt.typing import REAL
from pysmt.walkers import IdentityDagWalker

from wmipa.utils import is_atom


class WeightConverter(ABC):
    def __init__(self, variables):
        self.variables = variables

    @abstractmethod
    def convert(self, weight_func):
        """Convert the weight function into a LRA formula

        Args:
            weight_func (FNode): The weight function

        Returns:
            FNode: The formula representing the weight function
        """
        pass


class WeightConverterEUF(WeightConverter):
    def __init__(self, variables):
        super().__init__(variables)
        self.conv_aliases = set()
        self.prod_fn = FreshSymbol(
            typename=FunctionType(REAL, (REAL, REAL)), template="PROD%s"
        )
        self.pow_fn = FreshSymbol(
            typename=FunctionType(REAL, (REAL, REAL)), template="POW%s"
        )
        self.exp_fn = FreshSymbol(
            typename=FunctionType(REAL, (REAL,)), template="EXP%s"
        )
        self.counter = 1

    def convert(self, weight_func):
        conversion_list = list()
        w, _ = self.convert_rec(weight_func, Bool(False), conversion_list)
        if not w.is_symbol():
            y = self.variables.new_weight_alias(len(self.conv_aliases))
            conversion_list.append(Equals(y, w))

        return And(conversion_list)

    def convert_rec(self, formula, branch_condition, conversion_list):

        if formula.is_ite():
            return self._process_ite(formula, branch_condition, conversion_list)
        elif formula.is_times():
            return self._process_times(formula, branch_condition, conversion_list)
        elif formula.is_pow():
            return self._process_pow(formula, branch_condition, conversion_list)
        elif formula.is_exp():
            return self._process_exp(formula, branch_condition, conversion_list)
        elif len(formula.args()) == 0:
            return formula, False
        else:
            has_cond = False
            new_children = []
            for arg in formula.args():
                child, cond = self.convert_rec(arg, branch_condition, conversion_list)
                has_cond = has_cond or cond
                new_children.append(child)
            return (
                get_env().formula_manager.create_node(
                    node_type=formula.node_type(), args=tuple(new_children)
                ),
                has_cond,
            )

    def _process_ite(self, formula, branch_condition, conversion_list):
        phi, left, right = formula.args()
        # update branch conditions for children and recursively convert them
        l_cond = Or(branch_condition, Not(phi))
        r_cond = Or(branch_condition, phi)
        left, l_has_cond = self.convert_rec(left, l_cond, conversion_list)
        right, r_has_cond = self.convert_rec(right, r_cond, conversion_list)

        if not l_has_cond:
            left = self.variables.new_weight_alias(len(self.conv_aliases))
            f = self.variables.new_EUF_alias(self.counter)
            e = Equals(left, f)
            self.counter += 1
            conversion_list.append(Or(branch_condition, e))
        if not r_has_cond:
            right = self.variables.new_weight_alias(len(self.conv_aliases))
            f = self.variables.new_EUF_alias(self.counter)
            e = Equals(right, f)
            self.counter += 1
            conversion_list.append(Or(branch_condition, e))

        # add     (branch_conditions) -> (y = left) AND
        #     (not branch_conditions) -> (y = right)
        y = self.variables.new_weight_alias(len(self.conv_aliases))
        self.conv_aliases.add(y)
        el = Equals(y, left)
        er = Equals(y, right)
        conversion_list.append(Or(Or(branch_condition, Not(phi)), el))
        conversion_list.append(Or(Or(branch_condition, phi), er))
        # add      (branch_condition) -> not (y = left) OR not (y = right)
        # to force the enumeration of relevant branch conditions
        conversion_list.append(Or(branch_condition, Not(el), Not(er)))
        return y, True

    def _process_times(self, formula, branch_condition, conversion_list):
        args = (
            self.convert_rec(arg, branch_condition, conversion_list)
            for arg in formula.args()
        )
        const_val = 1
        others = []
        has_cond = False
        for arg, cond in args:
            has_cond = has_cond or cond
            if arg.is_real_constant():
                const_val *= arg.constant_value()
            else:
                others.append(arg)
        const_val = Real(const_val)
        if not others:
            return const_val, has_cond
        else:
            # abstract product between non-constant nodes with EUF
            return Times(const_val, reduce(self.prod_fn, others)), has_cond

    def _process_pow(self, formula, branch_condition, conversion_list):
        (base, b_has_cond) = self.convert_rec(formula.arg(0), branch_condition, conversion_list)
        (exponent, e_has_cond) = self.convert_rec(formula.arg(1), branch_condition, conversion_list)

        has_cond = b_has_cond or e_has_cond
        if exponent.is_zero():
            return Real(1), has_cond
        else:
            return self.pow_fn(base, exponent), has_cond

    def _process_exp(self, formula, branch_condition, conversion_list):
        exponent, has_cond = self.convert_rec(formula.arg(0), branch_condition, conversion_list)
        if exponent.is_zero():
            return Real(1), has_cond
        else:
            # expand power into product and abstract it with EUF
            return self.exp_fn(exponent), has_cond


class WeightConverterSkeleton(WeightConverter):
    def __init__(self, variables):
        super().__init__(variables)
        self.cnf_labels = set()
        self.cnfizer = LabelCNFizer(self.variables)

    def new_label(self):
        B = self.variables.new_cond_label(len(self.cnf_labels))
        self.cnf_labels.add(B)
        return B

    def convert(self, weight_func):
        conversion_list = list()
        self._convert_rec(weight_func, Bool(False), conversion_list)
        return And(conversion_list)

    def _convert_rec(self, formula, branch_condition, conversion_list):
        if formula.is_ite():
            return self._process_ite(formula, branch_condition, conversion_list)
        elif formula.is_theory_op():
            for arg in formula.args():
                self._convert_rec(arg, branch_condition, conversion_list)

    def _process_ite(self, formula, branch_condition, conversion_list):
        phi, left, right = formula.args()
        if is_atom(phi):
            if branch_condition is not None:
                l_cond = Or(branch_condition, Not(phi))
                r_cond = Or(branch_condition, phi)
            else:
                l_cond = Not(phi)
                r_cond = phi
            conversion_list.append(Or(branch_condition, phi, Not(phi)))
            self._convert_rec(left, l_cond, conversion_list)
            self._convert_rec(right, r_cond, conversion_list)
        else:
            b = self.new_label()
            if branch_condition is not None:
                l_cond = Or(branch_condition, Not(b))
                r_cond = Or(branch_condition, b)
            else:
                l_cond = Not(b)
                r_cond = b
            for clause in self.cnfizer.convert(nnf(phi)):
                conversion_list.append(Or(l_cond, *clause))

            for clause in self.cnfizer.convert(nnf(Not(phi))):
                conversion_list.append(Or(r_cond, *clause))

            conversion_list.append(Or(branch_condition, b, Not(b)))

            self._convert_rec(left, l_cond, conversion_list)
            self._convert_rec(right, r_cond, conversion_list)


class CNFPreprocessor(IdentityDagWalker):
    """
    Convert nested ORs and ANDs into flat lists of ORs and ANDs.
    """

    def walk_or(self, formula, args, **kwargs):
        children = []
        for arg in args:
            if arg.is_true():
                return self.mgr.Bool(True)
            elif arg.is_false():
                continue
            elif arg.is_or():
                children.extend(arg.args())
            elif arg.is_not() and arg.arg(0).is_and():
                children.extend(map(self.mgr.Not, arg.arg(0).args()))
            else:
                children.append(arg)
        return self.mgr.Or(children)

    def walk_and(self, formula, args, **kwargs):
        children = []
        for arg in args:
            if arg.is_false():
                return self.mgr.Bool(False)
            elif arg.is_true():
                continue
            elif arg.is_and():
                children.extend(arg.args())
            elif arg.is_not() and arg.arg(0).is_or():
                children.extend(map(self.mgr.Not, arg.arg(0).args()))
            else:
                children.append(arg)
        return self.mgr.And(children)

    def walk_implies(self, formula, args, **kwargs):
        left, right = formula.args()
        left_a, right_a = args
        return self.walk_or(self.mgr.Or(self.mgr.Not(left), right), (self.mgr.Not(left_a), right_a), **kwargs)


class LabelCNFizer(CNFizer):
    def walk_quantifier(self, formula, args, **kwargs):
        pass

    def __init__(self, wmi_variables, environment=None):
        super().__init__(environment)
        self.preprocessor = CNFPreprocessor(env=environment)
        self.wmi_variables = wmi_variables

    def _key_var(self, formula):
        if formula in self._introduced_variables:
            res = self._introduced_variables[formula]
        else:
            res = self.wmi_variables.new_cnf_label(len(self._introduced_variables))
            self._introduced_variables[formula] = res
        return res

    def convert(self, formula):
        """Convert formula into an equisatisfiable CNF.

        Returns a set of clauses: a set of sets of literals.
        """

        formula = self.preprocessor.walk(formula)
        tl, _cnf = self.walk(formula)

        if not _cnf:
            return [frozenset([tl])]
        res = []
        for clause in _cnf:
            if len(clause) == 0:
                return CNFizer.FALSE_CNF
            simp = []
            for lit in clause:
                if lit is tl or lit.is_true():
                    # Prune clauses that are trivially TRUE
                    # and clauses containing the top level label
                    simp = None
                    break
                elif not lit.is_false() and lit is not Not(tl):
                    # Prune FALSE literals
                    simp.append(lit)
            if simp:
                res.append(frozenset(simp))

        return frozenset(res)

    def walk_not(self, formula, args, **kwargs):
        a, _cnf = args[0]
        if a.is_true():
            return self.mgr.Bool(False), CNFizer.TRUE_CNF
        elif a.is_false():
            return self.mgr.Bool(True), CNFizer.TRUE_CNF
        else:
            return Not(a), _cnf

    def walk_and(self, formula, args, **kwargs):
        if len(args) == 1:
            return args[0]

        k = self._key_var(formula)

        _cnf = []
        for a, c in args:
            _cnf.append(frozenset([a, self.mgr.Not(k)]))
            for clause in c:
                _cnf.append(clause)

        return k, frozenset(_cnf)

    def walk_or(self, formula, args, **kwargs):
        if len(args) == 1:
            return args[0]
        k = self._key_var(formula)

        _cnf = [frozenset([self.mgr.Not(k)] + [a for a, _ in args])]
        for a, c in args:
            for clause in c:
                _cnf.append(clause)

        return k, frozenset(_cnf)


class SkeletonSimplifier(Simplifier):
    """Simplifier that does not simplify formulas like Or(phi, Not(phi))"""

    def walk_or(self, formula, args, **kwargs):
        if len(args) == 2 and args[0] == args[1]:
            return args[0]

        new_args = set()
        for a in args:
            if a.is_false():
                continue
            if a.is_true():
                return self.manager.TRUE()
            if a.is_or():
                for s in a.args():
                    new_args.add(s)
            else:
                new_args.add(a)

        if len(new_args) == 0:
            return self.manager.FALSE()
        elif len(new_args) == 1:
            return next(iter(new_args))
        else:
            return self.manager.Or(new_args)



"""This module implements the utility classes used to convert LRA expressions,
such as linear inequalities and polynomial weight functions, into a format that
is more suitable to be processed by LattE Integrale.

"""

__version__ = "0.99"
__author__ = "Paolo Morettin"

from pysmt.shortcuts import LT, And, Bool, Plus, Real, Symbol, Times
from pysmt.typing import REAL

from wmipa.integration.polynomial import Polynomial
from wmipa.utils import lcmm
from wmipa.wmiexception import WMIParsingException


class Bound:
    """Intermediate representation of a linear inequality.

    The inequality is rescaled in order to have integer coefficients.

    (c_1 * x_1 + ... + c_k * x_k) OP real_const
        where OP = {<, <=, >, >=}.

    Attributes:
        constant (int): int(real_const * LCD).
        coefficients (dict): Dictionary {x_i : int(c_i * LCD)}.

    where LCD = LCM(denominators({c_1, ... , c_k, real_const})).

    """

    def __init__(self, expression, aliases=None):
        """Default constructor.

        Takes as input a pysmt formula representing a linear inequality and
            a dictionary of aliases to be substituted before the parsing.

        Args:
            expression (FNode): The pysmt formula representing the inequality.
            aliases (dict {FNode : FNode}): The dictionary containing the aliases
                definitions.

        Raises:
            WMIParsingException: If the expression is not an inequality or the
                polynomial has degree more than 1.

        """
        if aliases is None:
            aliases = {}
        if not (expression.is_le() or expression.is_lt()):
            raise WMIParsingException(WMIParsingException.NOT_AN_INEQUALITY, expression)
        left, right = expression.args()
        if right.is_real_constant():
            # Polynomial OP Constant
            self._parse_expression(left, right, False, aliases)
        elif left.is_real_constant():
            # Constant OP Polynomial
            self._parse_expression(right, left, True, aliases)
        else:
            # Polynomial1 OP Polynomial2  converted into  Polynomial1 - Polynomial2 OP 0
            self._parse_expression(
                Plus(left, Times(Real(-1), right)), Real(0), False, aliases
            )

    def _parse_expression(self, polynomial, constant, negate, aliases):
        """Parse the expression representing the inequality.

        Args:
            polynomial (FNode): The pysmt formula representing the polynomial part.
            constant (FNode): The pysmt formula representing the constant part.
            negate (bool): If True the inequality will be flipped.
            aliases (dict {FNode : FNode}): The dictionary containing the aliases
                definitions.

        Raises:
            WMIParsingException: If the polynomial has degree more than 1.

        """
        assert constant.is_real_constant(), "Not a Real instance: " + str(constant)
        b = constant.constant_value()
        poly = Polynomial(polynomial, aliases)

        if negate:
            poly.negate()
            b = -b

        if poly.degree() > 1:
            raise WMIParsingException(
                WMIParsingException.POLYNOMIAL_WITH_DEGREE_GREATER_THAN_ONE, poly
            )

        self.coefficients = {}
        self.constant = b
        if poly.degree() > 0:
            # After the alias substitutions, the polynomial may contain monomials of
            # degree 0, which must be moved to the constant part
            variable_monomials = []
            for monomial in poly.monomials:
                if monomial.degree() == 0:
                    b -= monomial.coefficient
                else:
                    variable_monomials.append(monomial)
            poly.monomials = variable_monomials

            # Convert the constants to integers (LattE requirement for the polytope)
            denominators = [m.coefficient.denominator for m in poly.monomials] + [
                b.denominator
            ]
            lcd = lcmm(denominators)
            for monomial in poly.monomials:
                assert (
                        len(monomial.exponents) == 1
                        and list(monomial.exponents.values())[0] == 1
                ), "Not a monomial of degree 1"
                name = list(monomial.exponents.keys())[0]
                assert name not in self.coefficients, "Polynomial not in canonical form"
                self.coefficients[name] = int(monomial.coefficient * lcd)
            self.constant = int(b * lcd)
        else:
            pass
            # self.constant = Minus(Real(b), polynomial)

    def __str__(self):
        """The str method.

        Returns:
            str: The string representation of the inequality (e.g: (3x) + (-2y) < 7).

        """
        if len(self.coefficients) > 0:
            polynomial = " + ".join(
                ["(" + str(c) + " * " + x + ")" for x, c in self.coefficients.items()]
            )
            return polynomial + " < " + str(self.constant)
        else:
            return "0 < " + str(self.constant)

    def __eq__(self, other):
        if isinstance(other, Bound):
            return (
                    self.constant == other.constant
                    and self.coefficients == other.coefficients
            )
        return False

    def __hash__(self):
        return hash(self.constant)

    def to_pysmt(self):
        if not self.coefficients:
            return Bool(True)
        polynomial = Plus(
            *(Times(Real(c), Symbol(x, REAL)) for x, c in self.coefficients.items()),
        )
        return LT(polynomial, Real(self.constant))


class Polytope:
    """Intermediate representation of a polytope.

    Attributes:
        bounds (list(Bounds)): The list of all the inequalities in the polytope.
        variables (list): The list of the involved variables.

    """

    def __init__(self, expressions, aliases=None):
        """Default constructor.

        Takes as input a list of pysmt formulas representing the linear
            inequalities that define the polytope.

        Args:
            expressions (list(FNode)): The list of pysmt formulas representing the
                inequalities.
            aliases (dict {FNode : FNode}): The dictionary containing the aliases
                definitions.

        Raises:
            WMIParsingException: If one of the expressions is not an inequality or
                its polynomial has degree > 1.

        """
        if aliases is None:
            aliases = {}
        self.bounds = []
        for expr in expressions:
            b = Bound(expr, aliases)
            # After performing the aliases substitutions, the polynomials may be
            # simplified and have degree 0
            # In this case, ignore the inequality.
            if b.coefficients != {}:
                self.bounds.append(b)

        self.variables = set()
        for bound in self.bounds:
            for name in bound.coefficients.keys():
                self.variables.add(name)

    def __str__(self):
        """The str method.

        Returns:
            str: The string representation of the polytope
                (e.g: [(3x) + (-2y) < 7], [(5z < 10]).

        """
        bounds = ["[" + str(b) + "]" for b in self.bounds]
        return ", ".join(bounds)

    def is_empty(self):
        for i1 in range(len(self.bounds) - 1):
            for i2 in range(i1, len(self.bounds)):
                if self.bounds[i1].constant == -self.bounds[i2].constant:
                    coeff1 = self.bounds[i1].coefficients
                    coeff2 = self.bounds[i2].coefficients
                    if coeff1.keys() == coeff2.keys():
                        if all(coeff1[v] == -coeff2[v] for v in coeff1.keys()):
                            return True

        return False

    def to_pysmt(self):
        if not self.bounds:
            return Bool(True)
        return And(*map(lambda x: x.to_pysmt(), self.bounds))



from pysmt.shortcuts import FreshSymbol
from pysmt.typing import BOOL, REAL

from wmipa.utils import get_boolean_variables

COND = "cond"
WMI = "wmi"
QUERY = "query"
WEIGHT_ALIAS = "weight_alias"
CNF_LABEL = "weight_bool"
EUF_ALIAS = "euf_alias"


class WMIVariables:
    """This class handles all the variables that the program will create in order to calculate the WMI.

    Attributes:
        variables (dict): List of all the variables with relative type and index.

    """

    def __init__(self):
        """Default constructor"""
        self.variables = {}

    def new_cond_label(self, index):
        """Returns a symbol representing a condition label.

        Args:
            index (int): The index to associate to the label.

        Returns:
            FNode: The new label.

        """
        return self._new_label(COND, index)

    def new_query_label(self, index):
        """Returns a symbol representing a query label.

        Args:
            index (int): The index to associate to the label.

        Returns:
            FNode: The new label.

        """
        return self._new_label(QUERY, index)

    def new_wmi_label(self, index):
        """Returns a symbol representing a wmi label.

        Args:
            index (int): The index to associate to the label.

        Returns:
            FNode: The new label.

        """
        return self._new_label(WMI, index)

    def new_weight_alias(self, index):
        """Returns a symbol representing a weight label.

        Args:
            index (int): The index to associate to the label.

        Returns:
            FNode: The new label.

        """
        return self._new_label(WEIGHT_ALIAS, index, sym_type=REAL, template="FR%s")

    def new_EUF_alias(self, index):
        """Returns a symbol representing a weight label.

        Args:
            index (int): The index to associate to the label.

        Returns:
            FNode: The new label.

        """
        return self._new_label(EUF_ALIAS, index, sym_type=REAL, template="EUF%s")


    def new_cnf_label(self, index):
        """Returns a symbol representing a weight bool.

        Args:
            index (int): The index to associate to the bool.

        Returns:
            FNode: The new bool.

        """
        return self._new_label(CNF_LABEL, index, sym_type=BOOL, template="CNFB%s")

    def is_cond_label(self, variable):
        """Checks if the variable is a condition label.

        To recognize if the label is a condition label, it first checks if it is a wmi variable,
        then it controls its label type.

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a condition label, False otherwise.

        """
        return variable in self.variables and self.variables[variable][1] == COND

    def is_query_label(self, variable):
        """Checks if the variable is a query label.

        To recognize if the label is a query label, it first checks if it is a wmi variable,
        then it controls its label type.

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a query label, False otherwise.

        """
        return variable in self.variables and self.variables[variable][1] == QUERY

    def is_wmi_label(self, variable):
        """Checks if the variable is a wmi label.

        To recognize if the label is a wmi label, it first checks if it is a wmi variable,
        then it controls its label type.

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a wmi label, False otherwise.

        """
        return variable in self.variables and self.variables[variable][1] == WMI

    def is_euf_alias(self, variable):
        """Checks if the variable is a weight label.

        To recognize if the label is a weight label, it first checks if it is a weight variable,
        then it controls its label type.

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a weight label, False otherwise.

        """
        return variable in self.variables and self.variables[variable][1] == EUF_ALIAS

    def is_weight_alias(self, variable):
        """Checks if the variable is a weight label.

        To recognize if the label is a weight label, it first checks if it is a weight variable,
        then it controls its label type.

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a weight label, False otherwise.

        """
        return (
                variable in self.variables and self.variables[variable][1] == WEIGHT_ALIAS
        )

    def is_cnf_label(self, variable):
        """Checks if the variable is a weight bool.

        To recognize if the bool is a weight bool, it first checks if it is a weight variable,
        then it controls its label type.

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a weight bool, False otherwise.

        """
        return variable in self.variables and self.variables[variable][1] == CNF_LABEL

    def is_label(self, variable):
        """Checks if the variable is a condition label, query label or wmi label.

        To recognize if the label is of one of these types, this method checks if it is a wmi variable

        Args:
            variable (FNode): The variable to examine.

        Returns:
            bool: True if the variable is a condition, query or wmi label, False otherwise.

        """
        return variable in self.variables

    def contains_labels(self, formula):
        """Checks if the formula contains at least one boolean variable with reserved names.

        The possible reserved names are condition label, query label or wmi label.

        Args:
            formula (FNode): The pysmt formula to examine.

        Returns:
            bool: True if the formula contains at least one boolean variable with reserved names, False otherwise.

        """
        for var in get_boolean_variables(formula):
            if self.is_label(var):
                return True
        return False

    def _new_label(self, type_, index, sym_type=BOOL, template=None):
        """Create a new label based on the type and the index.

        Args:
            type_ (str): The type of the new label.
            index (int): The index to associate to the label.

        Returns:
            FNode: The new label.

        """
        symbol = FreshSymbol(typename=sym_type, template=template)
        assert symbol not in self.variables
        self.variables[symbol] = [index, type_]
        return symbol

    def get_label_index(self, variable):
        """Get the index associated to the variable.

        Args:
            variable (FNode): The variable to examine

        Returns:
            int: The index associated to the variable.

        """
        assert variable in self.variables
        return self.variables[variable][0]



"""This module implements some usefull methods used throughout the code.

Credits: least common multiple code by J.F. Sebastian
    (http://stackoverflow.com/a/147539)

"""

from functools import reduce

import networkx as nx
from pysmt.operators import POW
from pysmt.typing import BOOL, REAL

from wmipa.wmiexception import WMIParsingException


def get_boolean_variables(formula):
    """Finds all the boolean variables in the formula.

    Args:
        formula (FNode): The pysmt formula to examine.

    Returns:
        set(FNode): The set of all the boolean variables in the formula.

    """
    return _get_variables(formula, BOOL)


def get_real_variables(formula):
    """Finds all the real variables in the formula.

    Args:
        formula (FNode): The pysmt formula to examine.

    Returns:
        set(FNode): The set of all the real variables in the formula.

    """
    return _get_variables(formula, REAL)


def get_lra_atoms(formula):
    """Finds all the LRA atoms in the formula.

    Args:
        formula (FNode): The pysmt formula to examine.

    Returns:
        set(FNode): The set of all the LRA atoms in the formula.

    """
    return {a for a in formula.get_atoms() if a.is_theory_relation()}


def _get_variables(formula, type_):
    """Finds all the variables in the formula of the specific pysmt type.

    Args:
        formula (FNode): The pysmt formula to examine.
        type_: The pysmt type to find (e.g: REAL, BOOL).

    Returns:
        set(FNode): The set of all the variables in the formula of the specific type.

    """
    return {a for a in formula.get_free_variables() if a.get_type() == type_}


def is_pow(node):
    """Test whether the node is the Pow operator
        (this should be implemented in pysmt but is currently missing).

    Args:
        node (FNode): The node to examine.

    Returns:
        bool: True if the node is a Pow operator, False otherwise.

    """
    return node.node_type() == POW


def _gcd(a, b):
    """Computes the greatest common divisor of two numbers using Euclid's Algorithm.

    Example:
        >>> _gcd(30, 18)
        6

    Args:
        a (int): The first parameter.
        b (int): The second parameter.

    Returns:
        int: The greatest common divisor of a and b.

    """
    while b:
        a, b = b, a % b
    return a


def _lcm(a, b):
    """Compute the lowest common multiple of two numbers.

    Example:
        >>> _lcm(12, 20)
        60

    Args:
        a (int): The first parameter.
        b (int): The second parameter.

    Returns:
        int: The lowest common multiple of a and b.

    """
    return a * b // _gcd(a, b)


def lcmm(args):
    """Computes the lowest common multiple of a list of numbers.

    Example:
        >>> lcmm([5, 15, 12])
        60

    Args:
        args (list(int)): The list of numbers on which to compute the
            lowest common multiple.

    Returns:
        int: The lowest common multiple of all the numbers in the list.

    """
    return reduce(_lcm, args)

def apply_aliases(expression, aliases):
    """Substitute the aliases within the expression in the right order.

    Args:
        expression (FNode): The pysmt expression to apply the aliases to.
        aliases (dict): The aliases to apply to the expression.
    """
    if len(aliases) > 0:
        # Build a dependency graph of the substitutions and apply them in
        # topological order
        Gsub = nx.DiGraph()
        constant_subs = {}

        # For every alias
        for x, alias_expr in aliases.items():
            for y in alias_expr.get_free_variables():
                # Create a node from the alias to every symbol inside it
                Gsub.add_edge(x, y)
            # If the alias substitution leads to a constant value (e.g: PI = 3.1415)
            if len(alias_expr.get_free_variables()) == 0:
                constant_subs.update({x: alias_expr})

        # Get the nodes in topological order
        try:
            sorted_substitutions = [
                node for node in nx.topological_sort(Gsub) if node in aliases
            ]
        except nx.exception.NetworkXUnfeasible:
            raise WMIParsingException(
                WMIParsingException.CYCLIC_ASSIGNMENT_IN_ALIASES, aliases
            )

        # Apply all the substitutions
        for alias in sorted_substitutions:
            expression = expression.substitute({alias: aliases[alias]})
        expression = expression.substitute(constant_subs)
    return expression


def is_atom(node):
    return node.is_symbol(BOOL) or node.is_theory_relation()


def is_literal(node):
    return is_atom(node) or (node.is_not() and is_atom(node.arg(0)))


def is_clause(formula):
    return is_literal(formula) or (
            formula.is_or() and all(is_literal(l) for l in formula.args())
    )


def is_cnf(formula):
    return is_clause(formula) or (
            formula.is_and() and all(is_clause(c) for c in formula.args())
    )



from wmipa.integration.integrand import Integrand
from wmipa.utils import apply_aliases


class Expression(Integrand):
    """Generic Expression to integrate"""

    def __init__(self, expression, aliases=None):
        super().__init__()
        if aliases is None:
            aliases = {}

        self.expression = apply_aliases(expression, aliases)
        self.variables = set(map(str, self.expression.get_free_variables()))

    def __str__(self):
        return self.expression.serialize()

    def to_pysmt(self):
        return self.expression



from shutil import which

IMPORT_ERR_MSG = "No integration backend installed."
latte_installed = True
volesti_installed = True

# check whether LattE integrale is installed or not
if which("integrate") is None:
    latte_installed = False

# check whether VolEsti is installed or not
if which("volesti_integrate") is None:
    volesti_installed = False

if not any([latte_installed, volesti_installed]):
    raise ImportError(IMPORT_ERR_MSG)
else:
    from .integrator import Integrator
    from .cache_integrator import CacheIntegrator
    from .command_line_integrator import CommandLineIntegrator
    from .latte_integrator import LatteIntegrator
    from .volesti_integrator import VolestiIntegrator



from abc import ABC, abstractmethod


class Integrator(ABC):
    """This class represents the general integrator with which to compute integrals.

    For the moment there is only one integrator that uses LattE Integrale.

    """

    @abstractmethod
    def __init__(self, **options):
        """Default constructor.

        Args:
            **options: whatever option is needed for the integrator

        """
        pass

    @abstractmethod
    def integrate(self, atom_assignments, weight, aliases, *args, **kwargs):
        """Integrates a problem of the type {atom_assignments, weight, aliases}

        Args:
            problem (atom_assignments, weight, aliases): The problem to integrate.

        Returns:
            real: The integration result.

        """
        pass

    @abstractmethod
    def integrate_batch(self, problems, *args, **kwargs):
        """Integrates a batch of problems of the type {atom_assignments, weight, aliases}

        Args:
            problems (list(atom_assignments, weight, aliases)): The list of problems to
                integrate.

        """
        pass



"""This module implements the Weighted Model Integration calculation.

The calculation leverages:
    - a Satisfiability Modulo Theories solver supporting All-SMT (e.g. MathSAT)
    - a software computing exact volume of polynomials (e.g. LattE Integrale)

Currently, three algorithms are supported:
    "BC" -- The baseline Block-Clause method.
    "AllSMT" -- Improves BC by leveraging the AllSAT feature of the SMT Solver.
    "PA" -- WMI with Predicate Abstraction, may reduce drastically the number of
        integrals computed.
    "SA-PA" -- Improves PA by making it aware of the structure of the weight function.
    "SA-PA-SK" -- Improves SA-PA by using a better encoding of the problem and a revised
        enumeration strategy

"""

__version__ = "0.999"
__author__ = "Paolo Morettin"

import math
from collections import defaultdict
from math import fsum

import mathsat
from pysmt.shortcuts import (
    And,
    Bool,
    Iff,
    Implies,
    Not,
    Real,
    Solver,
    serialize,
    simplify,
    substitute,
)
from pysmt.typing import BOOL, REAL
from sympy import solve, sympify

from wmipa import logger
from wmipa.integration import LatteIntegrator
from wmipa.utils import get_boolean_variables, get_lra_atoms, get_real_variables
from wmipa.weightconverter import SkeletonSimplifier
from wmipa.weights import Weights
from wmipa.wmiexception import WMIParsingException, WMIRuntimeException
from wmipa.wmivariables import WMIVariables


class WMI:
    """The class that has the purpose to calculate the Weighted Module Integration of
        a given support, weight function and query.

    Attributes:
        variables (WMIVariables): The list of variables created and used by WMI.
        weights (Weights): The representation of the weight function.
        chi (FNode): The pysmt formula that contains the support of the formula
        integrator (Integrator): The integrator to use.

    """

    # WMI methods
    MODE_BC = "BC"
    MODE_ALLSMT = "AllSMT"
    MODE_PA = "PA"
    MODE_SA_PA = "SAPA"
    MODE_SA_PA_SK = "SAPASK"
    MODES = [MODE_BC, MODE_ALLSMT, MODE_PA, MODE_SA_PA, MODE_SA_PA_SK]

    def __init__(self, chi, weight=Real(1), integrator=None, **options):
        """Default constructor.

        Args:
            chi (FNode): The support of the problem.
            weight (FNode, optional): The weight of the problem (default: 1).
            integrator: integrator instance (default: LatteIntegrator()).

        """
        self.variables = WMIVariables()
        self.weights = Weights(weight, self.variables)
        self.chi = chi

        if integrator is None:
            integrator = LatteIntegrator()
        self.integrator = integrator

        self.list_norm = set()
        self.norm_aliases = dict()
        self.skeleton_simplifier = SkeletonSimplifier()

    def computeMI_batch(self, phis, **options):
        """Calculates the MI on a batch of queries.

        Args:
            phis (list(FNode)): The list of the queries on which to calculate the MI.
            **options:
                - domA: set of pysmt vars encoding the Boolean integration domain
                    (optional)
                - domX: set of pysmt vars encoding the real integration domain
                    (optional)
                - mode: The mode to use when calculating MI.

        Returns:
            list(real): The list containing the result of each computation.
            list(int): The list containing the number of integrations for each
                computation that have been computed.

        """
        # save the old weight
        old_weights = self.weights
        old_variables = self.variables

        # calculate wmi with constant weight
        new_variables = WMIVariables()
        new_weights = Weights(Real(1), new_variables)
        self.weights = new_weights
        self.variables = new_variables

        volumes, integrations = self.computeWMI_batch(phis, **options)

        # restore old weight
        self.weights = old_weights
        self.variables = old_variables
        return volumes, integrations

    def computeWMI_batch(self, phis, **options):
        """Calculates the WMI on a batch of queries.

        Args:
            phis (list(FNode)): The list of the queries on which to calculate the WMI.
            **options:
                - domA: set of pysmt vars encoding the Boolean integration domain
                    (optional)
                - domX: set of pysmt vars encoding the real integration domain
                    (optional)
                - mode: The mode to use when calculating WMI.

        Returns:
            list(real): The list containing the result of each computation.
            list(int): The list containing the number of integrations for each
                computation that have been computed.

        """
        volumes = []
        integrations = []

        for phi in phis:
            volume, n_integrations = self.computeWMI(phi, **options)
            volumes.append(volume)
            integrations.append(n_integrations)

        return volumes, integrations

    def computeMI(self, phi, **options):
        """Calculates the MI on a single query.

        Args:
            phi (FNode): The query on which to calculate the MI.
            **options:
                - domA: set of pysmt vars encoding the Boolean integration domain
                    (optional)
                - domX: set of pysmt vars encoding the real integration domain
                    (optional)
                - mode: The mode to use when calculating MI.

        Returns:
            real: The result of the computation.
            int: The number of integrations that have been computed.

        """
        # save old weight
        old_weights = self.weights
        old_variables = self.variables

        # calculate wmi with constant weight
        new_variables = WMIVariables()
        new_weights = Weights(Real(1), new_variables)
        self.weights = new_weights
        self.variables = new_variables

        volume, n_integrations = self.computeWMI(phi, **options)

        # restore old weight
        self.weights = old_weights
        self.variables = old_variables
        return volume, n_integrations

    def computeWMI(self, phi, **options):
        """Calculates the WMI on a single query.

        Args:
            phi (FNode): The query on which to calculate the WMI.
            **options:
                - domA: set of pysmt vars encoding the Boolean integration domain
                    (optional)
                - domX: set of pysmt vars encoding the real integration domain
                    (optional)
                - mode: The mode to use when calculating WMI.

        Returns:
            real: The result of the computation.
            int: The number of integrations that have been computed.

        """
        domA = options.get("domA")
        domX = options.get("domX")
        mode = options.get("mode")
        cache = options.get("cache")
        ete = options.get("ete")
        if mode is None:
            mode = WMI.MODE_PA
        if mode not in WMI.MODES:
            err = "{}, use one: {}".format(mode, ", ".join(WMI.MODES))
            logger.error(err)
            raise WMIRuntimeException(WMIRuntimeException.INVALID_MODE, err)
        if cache is None:
            cache = -1
        self.cache = cache

        formula = And(phi, self.chi)
        # formula = self.chi

        # Add the phi to the support
        if mode == WMI.MODE_SA_PA:
            formula = And(formula, self.weights.weights_as_formula_euf)
        elif "SK" in mode:
            formula = And(formula, self.weights.weights_as_formula_sk)
        else:
            formula = And(formula, self.weights.labelling)

        # print("FORMULA: ", serialize(formula))
        logger.debug("Computing WMI with mode: {}".format(mode))
        self.variables: WMIVariables
        x = {
            x
            for x in get_real_variables(formula)
            if not (self.variables.is_weight_alias(x) or self.variables.is_euf_alias(x))
        }
        A = {
            x for x in get_boolean_variables(formula) if not self.variables.is_label(x)
        }

        # Currently, domX has to be the set of real variables in the
        # formula, whereas domA can be a superset of the boolean
        # variables A. The resulting volume is multiplied by 2^|domA - A|.
        factor = 1
        logger.debug("A: {}, domA: {}".format(A, domA))
        if domA is not None:
            if len(A - domA) > 0:
                logger.error(
                    "Domain of integration mismatch: A - domA = {}".format(A - domA)
                )
                raise WMIRuntimeException(
                    WMIRuntimeException.DOMAIN_OF_INTEGRATION_MISMATCH, A - domA
                )
            else:
                factor = 2 ** len(domA - A)

        logger.debug("factor: {}".format(factor))
        if domX is not None and domX != x:
            logger.error("Domain of integration mismatch")
            raise WMIRuntimeException(
                WMIRuntimeException.DOMAIN_OF_INTEGRATION_MISMATCH, x - domX
            )

        if ete:
            formula = self._eager_theory_encoding(formula)

        compute_with_mode = {
            WMI.MODE_BC: self._compute_WMI_BC,
            WMI.MODE_ALLSMT: self._compute_WMI_AllSMT,
            WMI.MODE_PA: self._compute_WMI_PA,
            WMI.MODE_SA_PA: self._compute_WMI_SA_PA,
            #  WMI.MODE_SA_PA_BOOL: self._compute_WMI_SA_PA,
            #  WMI.MODE_SA_PA_BOOL_TA: self._compute_WMI_SA_PA_TA,
            #  WMI.MODE_SA_PA_BOOL_TA_TA: self._compute_WMI_SA_PA_TA_TA,
            WMI.MODE_SA_PA_SK: self._compute_WMI_SA_PA_SK,
        }

        volume, n_integrations, n_cached = compute_with_mode[mode](
            formula, self.weights
        )

        volume = volume * factor
        logger.debug(
            "Volume: {}, n_integrations: {}, n_cached: {}".format(
                volume, n_integrations, n_cached
            )
        )

        return volume, n_integrations

    def _eager_theory_encoding(self, chi):
        implications = []
        for var in get_real_variables(chi):
            imp_lt = {}
            imp_gt = {}
            for literal in set(get_lra_atoms(chi)):
                if get_real_variables(literal) == {var}:
                    # convert the inequality to 'var op const'
                    s = solve(sympify(serialize(literal)))
                    arg = 0
                    if "inf" in str(s.args[0]) or "oo" in str(s.args[0]):
                        arg = 1
                    dis = s.args[arg]
                    if dis.args[0].is_Number:
                        # n < var, n <= var
                        const = float(dis.args[0])
                        eq = 1 if literal.is_lt() else 0
                        d = imp_gt
                    else:
                        # var < n, var <= n
                        const = float(dis.args[1])
                        eq = 0 if literal.is_lt() else 1
                        d = imp_lt
                    key = (const, eq)
                    if key not in d:
                        d[key] = []
                    d[key].append(literal)
            imp_lt = sorted(list(imp_lt.items()))
            imp_gt = sorted(list(imp_gt.items()), reverse=True)

            def compare(key, other_key, index):
                if index == 0:
                    return key <= other_key
                elif index == 1:
                    return key >= other_key

            lists = [imp_lt, imp_gt]
            for index_l, l in enumerate(lists):
                other_l = lists[(index_l + 1) % 2]
                for i, (key, literals) in enumerate(l):

                    # all literals in 'literals' are the same => Iff
                    for j, literal in enumerate(literals):
                        if j < len(literals) - 1:
                            implications.append(Iff(literal, literals[j + 1]))

                    # literal at random in the list (because they are all the same)
                    literal = literals[0]
                    if i < len(l) - 1:
                        _, literals_next = l[i + 1]
                        implications.append(Implies(literal, literals_next[0]))

                    # implications between one list and another
                    if len(other_l) > 0:
                        other_key, _ = other_l[0]
                        if compare(key, other_key, index_l):
                            index = 0
                            while (
                                    compare(key, other_l[index][0], index_l)
                                    and index < len(other_l) - 1
                            ):
                                index += 1
                            # other list [index-1 was the last correct implication]
                            #   [1 = get list of literals]
                            #   [0 = get random literal]
                            other_literal = other_l[index - 1][1][0]
                            implications.append(Implies(literal, Not(other_literal)))
        implications = And(implications)
        return And(chi, implications)

    @staticmethod
    def check_consistency(formula):
        """Checks if the formula has at least a total truth assignment
            which is both theory-consistent and it propositionally satisfies it.

        Args:
            formula (FNode): The pysmt formula to examine.

        Returns:
            bool: True if the formula satisfies the requirements, False otherwise.

        """
        for _ in WMI._model_iterator_base(formula):
            return True

        return False

    @staticmethod
    def _model_iterator_base(formula):
        """Finds all the total truth assignments that satisfy the given formula.

        Args:
            formula (FNode): The pysmt formula to examine.

        Yields:
            model: The model representing the next total truth assignment that
                satisfies the formula.

        """
        solver = Solver(name="msat")
        solver.add_assertion(formula)
        while solver.solve():
            model = solver.get_model()
            yield model
            atom_assignments = {a: model.get_value(a) for a in formula.get_atoms()}

            # Constrain the solver to find a different assignment
            solver.add_assertion(
                Not(And([Iff(var, val) for var, val in atom_assignments.items()]))
            )

    @staticmethod
    def _callback(model, converter, result):
        """Callback method useful when performing AllSAT on a formula.

        This method takes the model, converts it into a more suitable form and finally
            adds it into the given results list.

        Args:
            model (list): The model created by the solver.
            converter: The class that converts the model.
            result (list): The list where to append the converted model.

        Returns:
            int: 1 (This method requires to return an integer)

        """
        py_model = [converter.back(v) for v in model]
        result.append(py_model)
        return 1

    def _compute_TTAs(self, formula):
        """Computes the total truth assignments of the given formula.

        This method first labels the formula and then uses the functionality of mathsat
            called AllSAT to retrieve all the total truth assignments.

        Args:
            formula (FNode): The pysmt formula to examine.

        Returns:
            list: The list of all the total truth assignments.
            dict: The dictionary containing all the correspondence between the labels
                and their true value.

        """
        labels = {}
        # expressions = []
        # allsat_variables = set()

        # Label LRA atoms with fresh boolean variables
        labelled_formula, pa_vars, labels = self.label_formula(
            formula, formula.get_atoms()
        )

        # Perform AllSMT on the labelled formula
        solver = Solver(name="msat")
        converter = solver.converter
        solver.add_assertion(labelled_formula)
        models = []
        mathsat.msat_all_sat(
            solver.msat_env(),
            [converter.convert(v) for v in pa_vars],
            lambda model: WMI._callback(model, converter, models),
        )
        return models, labels

    def _compute_WMI_AllSMT(self, formula, weights):
        """Computes WMI using the AllSMT algorithm.

        This method retrieves all the total truth assignments of the given formula,
            it then calculates the weight for the assignments and finally asks the
            integrator to compute the integral of the problem.

        Args:
            formula (FNode): The pysmt formula on which to compute the WMI.
            weights (FNode): The pysmt formula representing the weight function.

        Returns:
            real: The final volume of the integral computed by summing up all the
                integrals' results.
            int: The number of problems that have been computed.

        """
        models, labels = self._compute_TTAs(formula)
        problems = []
        for _, model in enumerate(models):
            # retrieve truth assignments for the original atoms of the formula
            atom_assignments = {}
            for atom, value in WMI._get_assignments(model).items():
                if atom in labels:
                    atom = labels[atom]
                atom_assignments[atom] = value
            problem = self._create_problem(atom_assignments, weights)
            problems.append(problem)

        results, cached = self.integrator.integrate_batch(problems, self.cache)
        volume = fsum(results)
        return volume, len(problems) - cached, cached

    def _create_problem(self, atom_assignments, weights, on_labels=True):
        """Create a tuple containing the problem to integrate.

        It first finds all the aliases in the atom_assignments, then it takes the
            actual weight (based on the assignment).
        Finally, it creates the problem tuple with all the info in it.

        Args:
            atom_assignments (dict): The list of assignments and relative value
                (True, False)
            weights (Weight): The weight of the problem.
            on_labels (bool): If True assignment is expected to be over labels of
                weight condition
                otherwise it is expected to be over unlabelled conditions

        Returns:
            tuple: The problem on which to calculate the integral formed by
                (atom assignment, actual weight, list of aliases).

        """
        aliases = {}
        for atom, value in atom_assignments.items():
            if value is True and atom.is_equals():
                alias, expr = self._parse_alias(atom)
                if self.variables.is_weight_alias(alias):
                    continue

                # check that there are no multiple assignments of the same alias
                if alias not in aliases:
                    aliases[alias] = expr
                else:
                    msg = "Multiple assignments to the same alias"
                    raise WMIParsingException(
                        WMIParsingException.MULTIPLE_ASSIGNMENT_SAME_ALIAS
                    )

        current_weight, cond_assignments = weights.weight_from_assignment(
            atom_assignments, on_labels=on_labels
        )
        return atom_assignments, current_weight, aliases, cond_assignments

    def _parse_alias(self, equality):
        """Takes an equality and parses it.

        Args:
            equality (FNode): The equality to parse.

        Returns:
            alias (FNode): The name of the alias.
            expr (FNode): The value of the alias.

        Raises:
            WMIParsingException: If the equality is not of the type
                (Symbol = real_formula) or vice-versa.

        """
        assert equality.is_equals(), "Not an equality"
        left, right = equality.args()
        if left.is_symbol() and (left.get_type() == REAL):
            alias, expr = left, right
        elif right.is_symbol() and (right.get_type() == REAL):
            alias, expr = right, left
        else:
            raise WMIParsingException(
                WMIParsingException.MALFORMED_ALIAS_EXPRESSION, equality
            )
        return alias, expr

    def _compute_WMI_BC(self, formula, weights):
        """Computes WMI using the Block Clause algorithm.

        This method retrieves all the total truth assignments of the given formula
            one by one thanks to the Block Clause algorithm.
            This particular algorithm finds a model that satisfy the formula, it then
            adds the negation of this model to the formula itself in order to constrain
            the solver to fine another model that satisfies it.

        Args:
            formula (FNode): The pysmt formula on which to compute the WMI.
            weights (FNode): The pysmt formula representing the weight function.

        Returns:
            real: The final volume of the integral computed by summing up all the
                integrals' results.
            int: The number of problems that have been computed.

        """
        problems = []
        for index, model in enumerate(WMI._model_iterator_base(formula)):
            atom_assignments = {
                a: model.get_value(a).constant_value() for a in formula.get_atoms()
            }
            problem = self._create_problem(atom_assignments, weights)
            problems.append(problem)

        results, cached = self.integrator.integrate_batch(problems, self.cache)
        volume = fsum(results)
        return volume, len(problems) - cached, cached

    def _compute_WMI_PA_no_boolean(
            self, lab_formula, pa_vars, labels, other_assignments=None
    ):
        """Finds all the assignments that satisfy the given formula using AllSAT.

        Args:
            lab_formula (FNode): The labelled pysmt formula to examine.
            pa_vars ():
            labels (dict): The dictionary containing the correlation between each label
                and their true value.

        Yields:
            dict: One of the assignments that satisfies the formula.

        """
        if other_assignments is None:
            other_assignments = {}
        solver = Solver(
            name="msat", solver_options={"dpll.allsat_minimize_model": "true"}
        )
        converter = solver.converter
        solver.add_assertion(lab_formula)
        lra_assignments = []
        mathsat.msat_all_sat(
            solver.msat_env(),
            [converter.convert(v) for v in pa_vars],
            lambda model: WMI._callback(model, converter, lra_assignments),
        )
        for mu_lra in lra_assignments:
            assignments = {}
            for atom, value in WMI._get_assignments(mu_lra).items():
                if atom in labels:
                    atom = labels[atom]
                assignments[atom] = value
            assignments.update(other_assignments)
            yield assignments

    def _compute_WMI_PA(self, formula, weights):
        """Computes WMI using the Predicate Abstraction (PA) algorithm.

        Args:
            formula (FNode): The formula on which to compute WMI.
            weights (Weight): The corresponding weight.

        Returns:
            real: The final volume of the integral computed by summing up all the
                integrals' results.
            int: The number of problems that have been computed.

        """
        problems = []
        boolean_variables = get_boolean_variables(formula)
        if len(boolean_variables) == 0:
            # Enumerate partial TA over theory atoms
            lab_formula, pa_vars, labels = self.label_formula(
                formula, formula.get_atoms()
            )
            # Predicate abstraction on LRA atoms with minimal models
            for assignments in self._compute_WMI_PA_no_boolean(
                    lab_formula, pa_vars, labels
            ):
                problem = self._create_problem(assignments, weights)
                problems.append(problem)
        else:
            solver = Solver(name="msat")
            converter = solver.converter
            solver.add_assertion(formula)
            boolean_models = []
            # perform AllSAT on the Boolean variables
            mathsat.msat_all_sat(
                solver.msat_env(),
                [converter.convert(v) for v in boolean_variables],
                lambda model: WMI._callback(model, converter, boolean_models),
            )

            logger.debug("n_boolean_models: {}".format(len(boolean_models)))
            # for each boolean assignment mu^A of F
            for model in boolean_models:
                atom_assignments = {}
                boolean_assignments = WMI._get_assignments(model)
                atom_assignments.update(boolean_assignments)
                subs = {k: Bool(v) for k, v in boolean_assignments.items()}
                f_next = formula
                # iteratively simplify F[A<-mu^A], getting (possibly part.) mu^LRA
                while True:
                    f_before = f_next
                    f_next = simplify(substitute(f_before, subs))
                    lra_assignments, over = WMI._parse_lra_formula(f_next)
                    subs = {k: Bool(v) for k, v in lra_assignments.items()}
                    atom_assignments.update(lra_assignments)
                    if over or lra_assignments == {}:
                        break

                if not over:
                    # predicate abstraction on LRA atoms with minimal models
                    lab_formula, pa_vars, labels = self.label_formula(
                        f_next, f_next.get_atoms()
                    )
                    expressions = []
                    for k, v in atom_assignments.items():
                        if k.is_theory_relation():
                            if v:
                                expressions.append(k)
                            else:
                                expressions.append(Not(k))

                    lab_formula = And([lab_formula] + expressions)
                    for assignments in self._compute_WMI_PA_no_boolean(
                            lab_formula, pa_vars, labels, atom_assignments
                    ):
                        problem = self._create_problem(assignments, weights)
                        problems.append(problem)
                else:
                    # integrate over mu^A & mu^LRA
                    problem = self._create_problem(atom_assignments, weights)
                    problems.append(problem)
        results, cached = self.integrator.integrate_batch(problems, self.cache)
        volume = fsum(results)
        return volume, len(problems) - cached, cached

    def _get_allsat(self, formula, use_ta=False, atoms=None, options=None):
        """
        Gets the list of assignments that satisfy the formula.

        Args:
            formula (FNode): The formula to satisfy
            use_ta (bool, optional): If true the assignments can be partial.
                Defaults to False.
            atoms (list, optional): List of atoms on which to find the assignments.
                Defaults to the boolean atoms of the formula.

        Yields:
            list: assignments on the atoms
        """
        if options is None:
            options = {}
        if use_ta:
            solver_options = {
                "dpll.allsat_minimize_model": "true",
                "dpll.allsat_allow_duplicates": "false",
                "preprocessor.toplevel_propagation": "false",
                "preprocessor.simplification": "0",
            }
        else:
            solver_options = {}

        solver_options.update(options)
        if atoms is None:
            atoms = get_boolean_variables(formula)

        for atom in atoms:
            if not atom.is_symbol(BOOL) and atom not in self.list_norm:
                _ = self.normalize_assignment(atom, True)
        solver = Solver(name="msat", solver_options=solver_options)
        converter = solver.converter

        solver.add_assertion(formula)

        models = []
        mathsat.msat_all_sat(
            solver.msat_env(),
            [converter.convert(v) for v in atoms],
            lambda model: WMI._callback(model, converter, models),
        )

        for model in models:
            assignments = {}
            for atom, value in WMI._get_assignments(model).items():
                if atom.is_symbol(BOOL):
                    assignments[atom] = value
                else:
                    subterms = self.normalize_assignment(atom, False)

                    assert (
                            frozenset(subterms) in self.norm_aliases
                    ), "Atom {}\n(subterms: {})\nnot found in\n{}".format(
                        atom.serialize(),
                        subterms,
                        "\n".join(str(x) for x in self.norm_aliases.keys()),
                    )
                    for element_of_normalization in self.norm_aliases[
                        frozenset(subterms)
                    ]:
                        assignments[element_of_normalization] = (
                            not value if element_of_normalization.is_lt() else value
                        )
            yield assignments

    def normalize_coefficient(self, val, normalizing_coefficient):
        midvalue = round(float(val) / abs(normalizing_coefficient), 10)
        if abs(midvalue - round(float(midvalue))) < 0.00000001:
            return round(float(midvalue))
        return midvalue

    def normalize_assignment(self, assignment, add_to_norms_aliases):
        self.list_norm.add(assignment)
        if not assignment.is_lt():
            list_terms = [(assignment.args()[0], 1), (assignment.args()[1], -1)]
        else:
            list_terms = [(assignment.args()[0], -1), (assignment.args()[1], 1)]
        symbols = defaultdict(float)
        normalizing_coefficient = 0.0
        min_abs_coefficient = math.inf

        for term, coeff in list_terms:
            if term.is_real_constant():
                normalizing_coefficient += term.constant_value() * coeff
            elif term.is_plus():
                list_terms.extend([(atom, coeff) for atom in term.args()])
            elif term.is_minus():
                child1 = term.args()[0]
                child2 = term.args()[1]
                list_terms.append((child1, coeff))
                list_terms.append((child2, -1.0 * coeff))
            elif term.is_times():
                child1 = term.args()[0]
                child2 = term.args()[1]

                if child1.is_real_constant() and child2.is_real_constant():
                    # constant += child1.constant_value() * child2.constant_value()
                    list_terms.append(
                        (child1.constant_value() * child2.constant_value(), coeff)
                    )
                elif child1.is_real_constant():
                    list_terms.append((child2, coeff * child1.constant_value()))
                elif child2.is_real_constant():
                    list_terms.append((child1, coeff * child2.constant_value()))
            else:
                symbols[term] += coeff
                min_abs_coefficient = min(min_abs_coefficient, abs(coeff))

        normalized_assignment = list()
        normalized_assignment2 = list()

        if assignment.is_equals():
            normalized_assignment.append("=")
            normalized_assignment2.append("=")
        else:
            normalized_assignment.append("<=/>=")

        if normalizing_coefficient == 0.0:
            normalized_assignment.append(0.0)
            normalizing_coefficient = min_abs_coefficient
            if assignment.is_equals():
                normalized_assignment2.append(0.0)
        else:
            normalized_assignment.append(
                normalizing_coefficient / abs(normalizing_coefficient)
            )
            if assignment.is_equals():
                normalized_assignment2.append(
                    (-normalizing_coefficient / abs(normalizing_coefficient))
                )

        for term in symbols:
            normalized_assignment.append(
                (
                    term,
                    self.normalize_coefficient(symbols[term], normalizing_coefficient),
                )
            )
            if assignment.is_equals():
                normalized_assignment2.append(
                    (
                        term,
                        self.normalize_coefficient(
                            -symbols[term], normalizing_coefficient
                        ),
                    )
                )

        if add_to_norms_aliases:
            if frozenset(normalized_assignment) not in self.norm_aliases:
                self.norm_aliases[frozenset(normalized_assignment)] = list()
            self.norm_aliases[frozenset(normalized_assignment)].append(assignment)
            # print("ADDING", frozenset(normalized_assignment), "for", assignment)
            if assignment.is_equals():
                if frozenset(normalized_assignment2) not in self.norm_aliases:
                    self.norm_aliases[frozenset(normalized_assignment2)] = list()
                self.norm_aliases[frozenset(normalized_assignment2)].append(assignment)
                # print("ADDING", frozenset(normalized_assignment2), "for", assignment)

        return None if add_to_norms_aliases else normalized_assignment

    def _compute_WMI_PA_no_boolean_no_label(self, lra_formula, other_assignments=None):
        """Finds all the assignments that satisfy the given formula using AllSAT.

        Args:
            lra_formula (FNode): The non-labelled pysmt formula to examine.

        Yields:
            dict: One of the assignments that satisfies the formula.

        """
        # for SAPA filter out new aliases for the weight function written as a formula
        if other_assignments is None:
            other_assignments = {}
        lra_atoms = {
            atom
            for atom in lra_formula.get_atoms()
            if not (self.variables.is_cnf_label(atom) or (
                    atom.is_equals() and self.variables.is_weight_alias(atom.arg(0))
            ))
        }
        # bools = {x for x in get_boolean_variables(lra_formula)
        #         if not self.variables.is_weight_bool(x)}
        # assert len(bools) == 0, bools

        lra_assignments = self._get_allsat(lra_formula, use_ta=True, atoms=lra_atoms)
        for mu_lra in lra_assignments:
            mu_lra.update(other_assignments)
            yield mu_lra

    def _simplify_formula(self, formula, subs, atom_assignments):
        """Substitute the subs in the formula and iteratively simplify it.
        atom_assignments is updated with unit-propagated atoms.

        Args:
            formula (FNode): The formula to simplify.
            subs (dict): Dictionary with the substitutions to perform.
            atom_assignments (dict): Dictionary with atoms and assigned value.

        Returns:
            bool: True if the formula is completely simplified.
            FNode: The simplified formula.
        """
        subs = {k: Bool(v) for k, v in subs.items()}
        f_next = formula
        # iteratively simplify F[A<-mu^A], getting (possibly part.) mu^LRA
        while True:
            f_before = f_next
            f_next = self.skeleton_simplifier.simplify(substitute(f_before, subs))
            lra_assignments, over = WMI._parse_lra_formula(f_next)
            subs = {k: Bool(v) for k, v in lra_assignments.items()}
            atom_assignments.update(lra_assignments)
            if over or lra_assignments == {}:
                break

        if not over:
            # formula not completely simplified, add conjunction of assigned LRA atoms
            expressions = []
            for k, v in atom_assignments.items():
                if k.is_theory_relation():
                    if v:
                        expressions.append(k)
                    else:
                        expressions.append(Not(k))
            f_next = And([f_next] + expressions)
        return over, f_next

    def _compute_WMI_SA_PA(self, formula, weights, use_ta=True):
        """Computes WMI using the Predicate Abstraction (PA) algorithm using Structure
            Awareness.

        Args:
            formula (FNode): The formula on which to compute WMI.
            weights (Weight): The corresponding weight.

        Returns:
            real: The final volume of the integral computed by summing up all the
                integrals' results.
            int: The number of problems that have been computed.

        """
        problems = []

        boolean_variables = get_boolean_variables(formula)

        # number of booleans not assigned in each problem
        n_bool_not_assigned = []
        if len(boolean_variables) == 0:
            # Enumerate partial TA over theory atoms
            for assignments in self._compute_WMI_PA_no_boolean_no_label(formula):
                problem = self._create_problem(assignments, weights, on_labels=False)
                problems.append(problem)
                n_bool_not_assigned.append(0)
        else:
            boolean_models = self._get_allsat(
                formula, use_ta=use_ta, atoms=boolean_variables
            )

            # logger.debug("n_boolean_models: {}".format(len(boolean_models)))
            # for each (partial) boolean assignment mu^A of F
            for boolean_assignments in boolean_models:
                atom_assignments = dict(boolean_assignments)
                # simplify the formula
                over, lra_formula = self._simplify_formula(
                    formula, boolean_assignments, atom_assignments
                )

                residual_booleans = get_boolean_variables(lra_formula)

                # if some boolean have not been simplified, find TTA on them
                if len(residual_booleans) > 0:
                    # compute TTA
                    residual_boolean_models = self._get_allsat(
                        lra_formula, atoms=residual_booleans
                    )
                else:
                    # all boolean variables have been assigned
                    residual_boolean_models = [[]]

                for residual_boolean_assignments in residual_boolean_models:
                    curr_atom_assignments = dict(atom_assignments)
                    if len(residual_boolean_assignments) > 0:
                        # simplify the formula
                        curr_atom_assignments.update(residual_boolean_assignments)
                        over, curr_lra_formula = self._simplify_formula(
                            lra_formula,
                            residual_boolean_assignments,
                            curr_atom_assignments,
                        )
                    else:
                        curr_lra_formula = lra_formula

                    b_not_assigned = len(boolean_variables) - len(boolean_assignments) - len(
                        residual_boolean_assignments)

                    if not over:
                        # predicate abstraction on LRA atoms with minimal models
                        for assignments in self._compute_WMI_PA_no_boolean_no_label(
                                curr_lra_formula, curr_atom_assignments
                        ):
                            problem = self._create_problem(
                                assignments, weights, on_labels=False
                            )
                            problems.append(problem)
                            n_bool_not_assigned.append(b_not_assigned)
                    else:
                        # integrate over mu^A & mu^LRA
                        problem = self._create_problem(
                            curr_atom_assignments, weights, on_labels=False
                        )

                        problems.append(problem)
                        n_bool_not_assigned.append(b_not_assigned)
        results, cached = self.integrator.integrate_batch(problems, self.cache)
        assert len(n_bool_not_assigned) == len(results)
        # multiply each volume by 2^(|A| - |mu^A|)
        volume = fsum(r * 2 ** i for i, r in zip(n_bool_not_assigned, results))
        return volume, len(problems) - cached, cached

    def _compute_WMI_SA_PA_SK(self, formula, weights):
        problems = []

        cnf_labels = {b for b in get_boolean_variables(formula) if
                      self.variables.is_cnf_label(b) or self.variables.is_cond_label(b)}
        boolean_variables = get_boolean_variables(formula) - cnf_labels
        lra_atoms = get_lra_atoms(formula)

        # number of booleans not assigned in each problem
        n_bool_not_assigned = []

        if len(boolean_variables) == 0:
            # Enumerate partial TA over theory atoms
            for assignments in self._get_allsat(formula, use_ta=True, atoms=lra_atoms):
                problem = self._create_problem(assignments, weights, on_labels=False)
                problems.append(problem)
                n_bool_not_assigned.append(0)

        else:
            boolean_models = self._get_allsat(
                formula, use_ta=True, atoms=boolean_variables,
            )

            for boolean_assignments in boolean_models:
                atom_assignments = dict(boolean_assignments)

                # simplify the formula
                over, res_formula = self._simplify_formula(
                    formula, boolean_assignments, atom_assignments
                )

                if not over:
                    # boolean variables first (discard cnf labels)
                    residual_atoms = list(get_boolean_variables(res_formula).intersection(boolean_variables)) + \
                                     list(get_lra_atoms(res_formula))

                    # may be both on LRA and boolean atoms
                    residual_models = self._get_allsat(
                        res_formula, use_ta=True, atoms=residual_atoms
                    )
                    for residual_assignments in residual_models:
                        curr_atom_assignments = dict(atom_assignments)
                        curr_atom_assignments.update(residual_assignments)

                        b_not_assigned = boolean_variables - curr_atom_assignments.keys()

                        problem = self._create_problem(
                            curr_atom_assignments, weights, on_labels=False
                        )
                        problems.append(problem)
                        n_bool_not_assigned.append(len(b_not_assigned))
                else:
                    b_not_assigned = boolean_variables - boolean_assignments.keys()

                    problem = self._create_problem(
                        atom_assignments, weights, on_labels=False
                    )
                    problems.append(problem)
                    n_bool_not_assigned.append(len(b_not_assigned))

        results, cached = self.integrator.integrate_batch(problems, self.cache)

        assert len(n_bool_not_assigned) == len(results)
        # multiply each volume by 2^(|A| - |mu^A|)
        volume = fsum(r * 2 ** i for i, r in zip(n_bool_not_assigned, results))
        return volume, len(problems) - cached, cached

    def label_formula(self, formula, atoms_to_label):
        """Labels every atom in the input with a new fresh WMI variable.

        Args:
            formula (FNode): The formula containing the atoms.
            atoms_to_label (list): The list of atoms to assign a new label.

        Returns:
            labelled_formula (FNode): The formula with the labels in it and their
                respective atoms.
            pa_vars (set): The list of all the atoms_to_label (as labels).
            labels (dict): The list of the labels and correspondent atom assigned to it.

        """
        expressions = []
        labels = {}
        pa_vars = set()
        j = 0
        for a in atoms_to_label:
            if a.is_theory_relation():
                label_a = self.variables.new_wmi_label(j)
                j += 1
                expressions.append(Iff(label_a, a))
                labels[label_a] = a
                pa_vars.add(label_a)
            else:
                pa_vars.add(a)

        labelled_formula = And([formula] + expressions)

        return labelled_formula, pa_vars, labels

    @staticmethod
    def _get_assignments(literals):
        """Retrieve the assignments (formula: truth value) from a list of literals
            (positive or negative).

        Args:
            literals (list): The list of the literals.

        Returns:
            assignments (dict): The list of atoms and corresponding truth value.

        """
        assignments = {}
        for literal in literals:
            if literal.is_not():
                value = False
                atom = literal.arg(0)
            else:
                value = True
                atom = literal
            assert atom.is_theory_relation or (
                    atom.is_symbol() and atom.get_type() == BOOL
            )
            assignments[atom] = value

        return assignments

    @staticmethod
    def _parse_lra_formula(formula):
        """Wrapper for _plra_rec.

        Args:
            formula (FNode): The formula to parse.

        Returns:
            dict: the list of FNode in the formula with the corresponding truth value.
            bool: boolean that indicates if there are no more truth assignment to
                extract.

        """
        return WMI._plra_rec(formula, True)

    @staticmethod
    def _plra_rec(formula, pos_polarity):
        """This method extract all sub formulas in the formula and returns them as a dictionary.

        Args:
            formula (FNode): The formula to parse.
            pos_polarity (bool): The polarity of the formula.

        Returns:
            dict: the list of FNode in the formula with the corresponding truth value.
            bool: boolean that indicates if there are no more truth assignment to
                extract.

        """
        if formula.is_bool_constant():
            return {}, True
        elif formula.is_theory_relation() or formula.is_symbol(BOOL):
            return {formula: pos_polarity}, True
        elif formula.is_not():
            return WMI._plra_rec(formula.arg(0), not pos_polarity)
        elif formula.is_and() and pos_polarity:
            assignments = {}
            over = True
            for a in formula.args():
                assignment, rec_over = WMI._plra_rec(a, True)
                assignments.update(assignment)
                over = rec_over and over
            return assignments, over
        elif formula.is_or() and not pos_polarity:
            assignments = {}
            over = True
            for a in formula.args():
                assignment, rec_over = WMI._plra_rec(a, False)
                assignments.update(assignment)
                over = rec_over and over
            return assignments, over
        elif formula.is_implies() and not pos_polarity:
            assignments, over_left = WMI._plra_rec(formula.arg(0), True)
            assignment_right, over_right = WMI._plra_rec(formula.arg(1), False)
            assignments.update(assignment_right)
            return assignments, over_left and over_right
        else:
            return {}, False



from itertools import product

from pysmt.fnode import FNode
from pysmt.shortcuts import And, Bool, Iff, get_env, serialize, simplify, substitute
from pysmt.typing import REAL

from wmipa.weightconverter import WeightConverterEUF, WeightConverterSkeleton
from wmipa.wmiexception import WMIParsingException


class Weights:
    """This class handles a FIUC weight function and provides a method that can evaluate
        the weight result given a truth assignment of all the conditions inside the
        weight function.

    Attributes:
        weights (FNode): The pysmt formula that represents the weight function.
        labels (set(FNode)): The set of all the condition labels created when labelling
            the weight function.
        labelling (FNode): The pysmt formula representing all the correlations between
            the labels and the actual conditions (e.g: cond_3 <-> (x < 5)).
        n_conditions (int): The number of all the conditions inside the weight function.
        cache (dict): A dictionary where each item is of the type
            {truth assignment : corresponding weight}

    """

    def __init__(self, weight_func, variables, cache=True, expand=False):
        """Initialize the weight function by labelling it with new condition labels and saves
            in a new formula the mapping from the labels to the relative condition.

        Args:
            weight_func (FNode): The pysmt formula representing the weight function.
            cache (bool, optional): The boolean that enables the use of the cache when
                evaluating the weight from a given assignment (default: True).
            expand (bool, optional): If True the cache will be populated with the values
                resulting from the application of the weight function on each possible
                truth assignment of the conditions inside it (default: False).

        """
        self.variables = variables

        # labels the weight function with new labels
        self.labelled_weights, subs = self.label_conditions(weight_func)
        self.labels = {}
        self.weight_conditions = {}
        for cond, sub in subs.items():
            self.labels[sub] = self.variables.get_label_index(sub)
            self.weight_conditions[cond] = self.variables.get_label_index(sub)
        self.n_conditions = len(subs)

        # create a formula representing all the substitution (e.g: cond_3 <-> (x < 5))
        labelling_list = []
        for cond, label in subs.items():
            labelling_list.append(Iff(cond, label))
        self.labelling = And(labelling_list)

        # convert weights as formula for EUF
        self.converterEUF = WeightConverterEUF(variables)
        self.converterSK = WeightConverterSkeleton(variables)
        self.weights_as_formula_euf = self.converterEUF.convert(weight_func)
        self.weights_as_formula_sk = self.converterSK.convert(weight_func)
        self.weights = weight_func

        # initialize the cache (if requested)
        if cache:
            self.cache = {}
            if expand:
                self._expand_to_dictionary()
        else:
            self.cache = None

    def weight_from_assignment(self, assignment, on_labels=True):
        """Given a total truth assignment of all the conditions inside the weight function,
            this method evaluates the resulting value based on the assignment.

        Args:
            assignment (dict {FNode : bool}): The dictionary containing all the
                assignments of each condition (e.g: {(x < 3) : True, (y < 5) : False}).
            on_labels (bool): If True assignment is expected to be over labels of weight
                condition otherwise it is expected to be over unlabelled conditions

        Returns:
            FNode: The result of the weight function on the given assignment.

        """
        if on_labels:
            conditions = self.labels
            weights = self.labelled_weights
        else:
            conditions = self.weight_conditions
            weights = self.weights

        # prepare an empty array of length 'n_conditions' to be populated with the
        # values of the assignment
        cond_assignment = [None for _ in range(self.n_conditions)]
        if on_labels:
            # populate the array with all the values
            for atom, value in assignment.items():
                assert isinstance(value, bool), "Assignment value should be Boolean"
                # if (atom.is_symbol() and atom.get_type() == BOOL and
                # self.variables.is_cond_label(atom)):
                if atom in conditions:
                    # take the index of the variable from the label
                    index = conditions[atom]
                    cond_assignment[index] = value

            assert (
                    None not in cond_assignment
            ), "Couldn't retrieve the complete assignment"

        assignment = {atom: Bool(v) for atom, v in assignment.items()}
        # evaluate the weight from the assignment
        if self.cache is not None:
            cond_assignment_tuple = tuple(cond_assignment)

            # first check into the cache if the weight associated to the particular
            # assignment was already calculated
            if cond_assignment_tuple in self.cache:
                return self.cache[cond_assignment_tuple], cond_assignment_tuple

            # else evaluate it and insert it into the cache
            else:
                flat_w = self._evaluate_weight(
                    weights, cond_assignment, conditions, assignment
                )
                cond_assignment_tuple = tuple(cond_assignment)
                self.cache[cond_assignment_tuple] = flat_w
                return flat_w, cond_assignment_tuple

        # if there is no cache just evaluate the weight
        else:
            flat_w = self._evaluate_weight(
                weights, cond_assignment, conditions, assignment
            )
            return flat_w, tuple(cond_assignment)

    def label_conditions(self, weight_func):
        """Finds and labels all the conditions inside the weight function with fresh boolean
            atoms.

        Args:
            weight_func (FNode): The pysmt formula representing the weight function to
                label.

        Returns:
            FNode: The pysmt formula representing the weight function with all the
                conditions substituted with the new labels.
            dict {FNode : FNode}: The dictionary containing all the conditions and their
                respectively substitution (e.g: {(x < 3) : cond_0, (y < 5) : cond_1} ).

        """
        # recursively find all the conditions and create a label for all of them
        subs = self._find_conditions(weight_func, subs={})

        # perform labelling
        labelled_weight_func = weight_func.substitute(subs)

        return labelled_weight_func, subs

    def _expand_to_dictionary(self):
        """Populates the cache with the values resulting from the application
        of the weight function on each possible truth assignment of
        the conditions
        inside it.

        """
        assert self.cache is not None, "Cache should be already initialized"

        # create every possible combination of assignments
        for assignment in product([True, False], repeat=self.n_conditions):
            # evaluate the results for all assignments and save them in the cache
            cond_w = self._evaluate_weight(
                self.labelled_weights, assignment, self.labels, {}
            )
            self.cache[assignment] = cond_w

    def _evaluate_weight(self, node, assignment, conditions, atom_assignment):
        """Evaluates the value of the given formula applied to the given assignment.

        Args:
            node (FNode): The pysmt formula representing (part of) the weight function.
            assignment (list(bool)): The dictionary containing the truth value of
                each assignment.

        Returns:
            FNode: The result of the formula representing (part of) the weight function
                applied to the given assignment.

        """
        # print("NODE", serialize(node), node.is_ite(), len(node.args()))
        # print()
        # print("CONDITIONS")
        # for c in conditions:
        if node.is_ite():
            cond, then, _else = node.args()
            # gets the index of the label to retrieve the truth value of that specific
            # label
            index_cond = conditions[cond]
            # assert assignment[index_cond] is not None, cond
            if assignment[index_cond] is None:
                assignment[index_cond] = self._evaluate_condition(cond, atom_assignment)
            # iteratively retrieve the leaf nodes from the 'then' or 'else' part,
            # depending on the truth value of the condition
            if assignment[index_cond]:
                return self._evaluate_weight(
                    then, assignment, conditions, atom_assignment
                )
            else:
                return self._evaluate_weight(
                    _else, assignment, conditions, atom_assignment
                )

        # found a leaf, return it
        elif len(node.args()) == 0:
            return node

        # this condition contains, for example, the Times operator
        else:
            new_children = []

            for child in node.args():
                new_children.append(
                    self._evaluate_weight(
                        child, assignment, conditions, atom_assignment
                    )
                )

            # after retrieving all the leaf nodes of the children it creates a new
            # node with these particular leafs
            return get_env().formula_manager.create_node(
                node_type=node.node_type(), args=tuple(new_children)
            )

    def _find_conditions(self, node, subs):
        """Finds all the conditions inside the given formula and creates a new label for each of
            them, adding both the condition and relative label to the given dictionary.

        Args:
            node (FNode): The pysmt formula representing (part of) the weight function.
            subs (dict {FNode : FNode}): The dictionary that will contain the
                correlations between each condition and their labels.

        Returns:
            dict {FNode : FNode}: The dictionary containing the correlations between
                each conditions of the given formula and their labels.

        Raises:
            WMIParsingException: If the weight formula does not represent a weight
                function.

        """
        if node.is_ite():
            cond, then, _else = node.args()

            # if this particular condition was not already labelled
            if cond not in subs:
                label = self.variables.new_cond_label(len(subs))
                subs[cond] = label

            # recursively finds all the conditions in both the 'then' and 'else' term
            subs = self._find_conditions(then, subs)
            subs = self._find_conditions(_else, subs)

        elif node.is_ira_op():
            # recursively finds all the conditions in all the children of the formula
            # (i.e: PLUS, MINUS, TIMES, DIV, POW)
            for child in node.args():
                subs = self._find_conditions(child, subs)
        elif len(node.args()) != 0 or not (node.is_symbol(REAL) or node.is_constant()):
            # other possible types of admitted nodes are symbol or constant else an
            # exception is raised
            raise WMIParsingException(WMIParsingException.INVALID_WEIGHT_FUNCTION, node)

        return subs

    def _evaluate_condition(self, condition, assignment):
        temp1 = condition.substitute(assignment)
        val = simplify(temp1)
        for _ in range(9):
            if val.is_bool_constant():
                break
            val = simplify(substitute(val, assignment))
        assert val.is_bool_constant(), (
                "Weight condition "
                + serialize(condition)
                + "\n\n cannot be evaluated with assignment "
                + "\n".join([str((x, assignment[x])) for x in assignment])
                + "\n\n simplified into "
                + serialize(val)
        )
        return val.constant_value()

    def __str__(self):
        ret = "Weight object {\n"
        ret += "\tlabelled weight: " + serialize(self.weights) + "\n"
        ret += "\tlabels: " + ", ".join([serialize(s) for s in self.labels]) + "\n"
        ret += "\tlabelling: " + serialize(self.labelling) + "\n"
        ret += "\tnumber of conditions: " + str(self.n_conditions) + "\n"
        ret += "}"

        return ret



import os
from os import path

from setuptools import find_packages, setup

NAME = "wmipa-test"
DESCRIPTION = "Weighted Model Integration PA (Predicate Abstraction) solver."
URL = "http://github.com/unitn-sml/wmi-pa"
EMAIL = "paolo.morettin@unitn.it"
AUTHOR = "Gabriele Masina, Paolo Morettin, Giuseppe Spallitta"
REQUIRES_PYTHON = ">=3.5.0"
VERSION = "0.1.8"

# What packages are required for this module to be executed?
REQUIRED = ["pysmt @ git+ssh://git@github.com/masinag/pysmt@nrat#egg=pysmt",
            "pywmi",
            "numpy",
            "sympy",
            "networkx"]

# What packages are optional?
EXTRAS = {
    #        'sdd': ["pysdd"]
}

here = os.path.abspath(os.path.dirname(__file__))

with open(path.join(here, "README.md")) as ref:
    long_description = ref.read()

"""
class PostInstallCommand(install):
    def run(self):
        install.run(self)
        os.system("pysmt-install --msat --confirm-agreement") # additionally install mathsat
"""

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=("test", "examples")),
    zip_safe=False,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    cmdclass={
        #        'upload': UploadCommand,
        #        'install': PostInstallCommand,
    },
)



__version__ = "0.999"
__author__ = "Gabriele Masina"

from subprocess import call

from wmipa.integration.command_line_integrator import CommandLineIntegrator
from wmipa.integration.expression import Expression
from wmipa.integration.polytope import Polytope
from wmipa.wmiexception import WMIRuntimeException


class VolestiIntegrator(CommandLineIntegrator):
    """This class handles the approximated integration of polynomial functions
    over (convex) polytopes.

    It inherits from the abstract class CacheIntegrator.

    VolEsti is required.

    Attributes:
        algorithm (str): The algorithm to use when computing the integrals.
        n_threads (int): The number of threads to use.
        error (int): The relative error tolerated (in (0.0, 1.0))
        walk_type (str): The type of random walk to use.
        N (int): Number of samples.
        walk_length (int): Length of random walk.

    """

    ALG_SEQUENCE_OF_BALLS = "SOB"
    ALG_COOLING_BALLS = "CB"
    ALG_COOLING_GAUSSIANS = "CG"
    DEF_ALGORITHM = ALG_COOLING_BALLS

    ALGORITHMS = [
        ALG_SEQUENCE_OF_BALLS,
        ALG_COOLING_BALLS,
        ALG_COOLING_GAUSSIANS,
    ]

    RW_BALL_WALK = "Ba"
    RW_RDHR = "RDHR"
    RW_CDHR = "CDHR"
    RW_BILLIARD_WALK = "Bi"
    RW_ACCELERATED_BILLIARD_WALK = "ABi"
    DEF_RANDOM_WALK = RW_CDHR

    RANDOM_WALKS = [
        RW_BALL_WALK,
        RW_RDHR,
        RW_CDHR,
        RW_BILLIARD_WALK,
        RW_ACCELERATED_BILLIARD_WALK,
    ]

    DEF_ERROR = 1e-1
    DEF_SEED = None  # unset
    DEF_WALK_LENGTH = None  # let the tool choose
    DEF_N = 1000

    def __init__(self,
                 algorithm=DEF_ALGORITHM,
                 error=DEF_ERROR,
                 walk_type=DEF_RANDOM_WALK,
                 walk_length=DEF_WALK_LENGTH,
                 seed=DEF_SEED,
                 N=DEF_N,
                 **options
                 ):
        """Default constructor.

        Args:
            algorithm: Defines the algorithm to use when integrating.
            error: The relative error tolerated (in (0.0, 1.0)).
            walk_type: The type of random walk to use.
            N: Number of samples.
            walk_length: Length of random walk.
            options: @see CommandLineIntegrator.__init__

        """
        CommandLineIntegrator.__init__(self, **options)

        self.algorithm = algorithm
        if self.algorithm not in self.ALGORITHMS:
            err = "{}, choose one from: {}".format(
                self.algorithm, ", ".join(self.ALGORITHMS)
            )
            raise WMIRuntimeException(WMIRuntimeException.INVALID_MODE, err)

        self.error = error
        if not (0.0 < self.error < 1.0):
            err = "{}, error must be in (0.0, 1.0)".format(self.error)
            raise WMIRuntimeException(WMIRuntimeException.OTHER_ERROR, err)

        self.walk_type = walk_type
        if self.walk_type not in self.RANDOM_WALKS:
            err = "{}, choose one from: {}".format(
                self.walk_type, ", ".join(self.RANDOM_WALKS)
            )
            raise WMIRuntimeException(WMIRuntimeException.INVALID_MODE, err)

        self.walk_length = walk_length
        if self.walk_length is not None and self.walk_length < 0:
            err = "{}, walk_length must be a non-negative number".format(
                self.walk_length
            )
            raise WMIRuntimeException(WMIRuntimeException.OTHER_ERROR, err)

        self.N = N
        if self.N <= 0:
            err = "{}, N must be a positive number".format(self.N)
            raise WMIRuntimeException(WMIRuntimeException.OTHER_ERROR, err)

        self.seed = seed
        if self.seed is not None and self.seed < 0:
            err = "{}, seed must be a non-negative number".format(self.seed)
            raise WMIRuntimeException(WMIRuntimeException.OTHER_ERROR, err)

    def _write_integrand_file(self, integrand, variables, path):
        """Writes the integrand to the given file.

        Args:
            integrand (Expression): The integrand.
            variables (list): The list of variables.
            path (str): The path where to write the integrand.

        """
        with open(path, "w") as f:
            f.write("{} \n {} \n".format(" ".join(variables), str(integrand)))

    @classmethod
    def _make_problem(cls, weight, bounds, aliases):
        """Makes the problem to be solved by VolEsti.
        Args:
            weight (FNode): The weight function.
            bounds (list): The polytope.
            aliases (dict): The aliases of the variables.
        Returns:
            integrand (Expression): The integrand.
            polytope (Polytope): The polytope.
        """
        integrand = Expression(weight, aliases)
        polytope = Polytope(bounds, aliases)
        return integrand, polytope

    def _call_integrator(self, polynomial_file, polytope_file, output_file):
        """Calls VolEsti executable to calculate the integrand of the problem
            represented by the given files.

        Args:
            polynomial_file (str): The path where to find the file with the
                representation of the polynomial.
            polytope_file (str): The path where to find the file with representation
                of the polytope.
            output_file (str): The file where to write the result of the computation.

        """
        cmd = [
            "volesti_integrate",
            polytope_file,
            polynomial_file,
            str(self.error),
            "--volume",
            self.algorithm,
            "--walk",
            self.walk_type,
            "--N",
            str(self.N),
        ]
        if self.walk_length is not None:
            cmd += ["--wlength", str(self.walk_length)]
        if self.seed is not None:
            cmd += ["--seed", str(self.seed)]

        with open(output_file, "w") as f:
            if self.stub_integrate:
                f.write("")
            else:
                return_value = call(cmd, stdout=f)



import re
import time
from abc import abstractmethod
from fractions import Fraction
from multiprocessing import Manager, Pool
from subprocess import call
from tempfile import NamedTemporaryFile

from pysmt.shortcuts import LE, LT

from wmipa.integration.integrator import Integrator
from wmipa.integration.polytope import Polynomial


class CacheIntegrator(Integrator):
    """This class handles the integration of polynomial functions over (convex) polytopes.
    It is a wrapper for an integrator that reads (writes) input (output) from (to) file.
    It implements different levels of caching (-1, 0, 1, 2, 3).

    It inherits from the abstract class Integrator.

    Attributes:
        n_threads (int): The number of threads to use.
        stub_integrate (bool): If True, the values will not be computed (0 is returned)
    """

    DEF_N_THREADS = 7

    def __init__(self, n_threads=DEF_N_THREADS, stub_integrate=False):
        """Default constructor.

        Args:
            **options:
                - algorithm: Defines the algorithm to use when integrating.
                - n_threads: Defines the number of threads to use.
                - stub_integrate: If True the integrals will not be computed

        """

        self.n_threads = n_threads
        self.stub_integrate = stub_integrate
        self.hashTable = HashTable()
        self.integration_time = 0.0

    def get_integration_time(self):
        return self.integration_time

    def _integrate_problem_or_cached(self, integrand, polytope, key, cache):
        value = self.hashTable.get(key)
        if value is not None:
            return value, True
        value = self._integrate_problem(integrand, polytope)
        if cache:
            self.hashTable.set(key, value)
        return value, False

    @abstractmethod
    def _integrate_problem(self, integrand, polytope) -> float:
        pass

    def cache_1(self, polytope, integrand, cond_assignments):
        variables = list(integrand.variables.union(polytope.variables))
        return self.hashTable.key(polytope, cond_assignments, variables), polytope

    def cache_2(self, polytope, integrand, _):
        return self.hashTable.key_2(polytope, integrand), polytope

    def cache_3(self, polytope, integrand, _):
        polytope = self._remove_redundancy(polytope)
        if polytope is None:
            return None, None
        return self.cache_2(polytope, integrand, _)

    def integrate_batch(self, problems, cache, *args, **kwargs):
        """Integrates a batch of problems of the type {atom_assignments, weight, aliases}

        Args:
            problems (list(atom_assignments, weight, aliases)): The list of problems to
                integrate.
            cache (int): The level of caching to use (range -1, 0, 1, 2, 3).

        """

        EMPTY = -1
        cache_modes = {
            -1: self.cache_1,
            0: self.cache_1,
            1: self.cache_1,
            2: self.cache_2,
            3: self.cache_3,
        }
        if cache not in cache_modes:
            modes = map(str, cache_modes.keys())
            raise Exception(
                f"Unsupported cache mode. Supported modes are ({', '.join(modes)})"
            )
        cache_fn = cache_modes[cache]

        problems_to_integrate = {}
        problem_id = []
        cached = 0
        for index, (atom_assignments, weight, aliases, cond_assignments) in enumerate(
                problems
        ):
            integrand, polytope = self._convert_to_problem(
                atom_assignments, weight, aliases
            )
            key, polytope = cache_fn(polytope, integrand, cond_assignments)
            if polytope is not None and not polytope.is_empty():
                # cache >= 1 recognize duplicates before calling the integrator
                pid = key if cache >= 1 else index
                if pid not in problems_to_integrate:
                    problem = (
                        len(problems_to_integrate),
                        integrand,
                        polytope,
                        key,
                        cache >= 0,
                    )
                    problems_to_integrate[pid] = problem
                else:
                    # duplicate found
                    cached += 1
                problem_id.append(problems_to_integrate[pid][0])
            else:
                problem_id.append(EMPTY)

        problems_to_integrate = problems_to_integrate.values()
        assert len(problem_id) == len(problems)
        # Handle multithreading
        start_time = time.time()
        pool = Pool(self.n_threads)
        results = pool.map(self._integrate_wrapper, problems_to_integrate)
        pool.close()
        pool.join()
        values = [0.0 if pid == EMPTY else results[pid][0] for pid in problem_id]
        cached += sum([(pid == EMPTY) or results[pid][1] for pid in problem_id])
        assert len(values) == len(problems)

        self.integration_time = time.time() - start_time
        return values, cached

    def _integrate_wrapper(self, problem):
        """A wrapper to handle multithreading."""
        _, integrand, polytope, key, cache = problem
        return self._integrate_problem_or_cached(integrand, polytope, key, cache)

    def integrate(self, atom_assignments, weight, aliases, cond_assignments, cache, *args, **kwargs):
        """Integrates a problem of the type {atom_assignments, weight, aliases}

        Args:
            problem (atom_assignments, weight, aliases): The problem to integrate.
            cond_assignments (tuple): truth values for the conditions
            cache (int): The level of caching to use (range -1, 0, 1, 2, 3).


        Returns:
            real: The integration result.

        """
        integrand, polytope = self._convert_to_problem(
            atom_assignments, weight, aliases
        )
        return self._integrate_problem_or_cached(
            integrand, polytope, cond_assignments, cache
        )

    @classmethod
    @abstractmethod
    def _make_problem(cls, weight, bounds, aliases):
        """Creates a problem of the type (integrand, polytope) from the given arguments.
        Args:
            weight (FNode): The weight of the integrand.
            bounds (list): The bounds of the polytope.
            aliases (dict): The aliases of the variables.

        Returns:
            integrand (Integrand): The problem to integrate.
            polytope (Polytope): The polytope to integrate over.
        """
        pass

    @classmethod
    def _convert_to_problem(cls, atom_assignments, weight, aliases):
        """Transforms an assignment into a problem, defined by:
            - a polynomial integrand
            - a convex polytope.

        Args:
            atom_assignments (dict): The assignment of the problem.
            weight (FNode): The weight of the problem.
            aliases (dict): The list of all the alias variables (like PI=3.14)

        Returns:
            integrand (Integrand): A representation of the weight.
            polytope (Polytope): The polytope representing the list of inequalities.

        """
        bounds = []
        for atom, value in atom_assignments.items():
            assert isinstance(value, bool), "Assignment value should be Boolean"

            # Skip atoms without variables
            if len(atom.get_free_variables()) == 0:
                continue

            if value is False:
                # If the negative literal is an inequality, change its
                # direction
                if atom.is_le():
                    left, right = atom.args()
                    atom = LT(right, left)
                elif atom.is_lt():
                    left, right = atom.args()
                    atom = LE(right, left)

            # Add a bound if the atom is an inequality
            if atom.is_le() or atom.is_lt():
                bounds.append(atom)

        return cls._make_problem(weight, bounds, aliases)

    def _remove_redundancy(self, polytope):
        polytope.bounds = list(set(polytope.bounds))
        polytope, internal_point = self._preprocess(polytope)
        if polytope is None:
            return None
        non_redundant_index = []
        to_analyze = list(range(0, len(polytope.bounds)))
        while len(to_analyze) > 0:
            index = to_analyze[0]
            non_redundant, essential_index = self._clarkson(
                polytope, internal_point, non_redundant_index, index
            )
            if non_redundant:
                non_redundant_index.append(essential_index)
            to_analyze.remove(essential_index)
        non_redundant_bounds = [polytope.bounds[i] for i in non_redundant_index]
        polytope.bounds = non_redundant_bounds
        return polytope

    def _preprocess(self, polytope):
        """
        maximize x_0
        subject to:
            Ax + 1x0 <= b
            x_0 <= 1

        x_0 is a new variable
        Ax <= b is the polytope
        """

        variables = list(polytope.variables)

        obj = [0] * len(variables) + [1]
        A = []
        b = []
        for bound in polytope.bounds:
            a = [0] * len(variables) + [1]
            for var_name in bound.coefficients:
                var_index = variables.index(var_name)
                a[var_index] = bound.coefficients[var_name]
            A.append(a)
            b.append(bound.constant)

        # x_0 <= 1
        A.append([0] * len(variables) + [1])
        b.append(1)

        optimal_value, optimal_solution = self._lp(A, b, obj, "maximize")

        # remove x_0
        optimal_solution = optimal_solution[:-1]

        if optimal_value > 0:
            # polytope dimension = len(variables)
            return polytope, optimal_solution
        elif optimal_value < 0:
            # polytope empty
            return None, []
        else:
            # polytope neither full-dimensional nor empty
            # TODO ?
            return polytope, optimal_solution

    def _clarkson(self, polytope, internal_point, non_redundant_index, index_to_check):
        """
        maximize A_k*x
        subject to:
            A_i*x <= b_i        for all i in I - k
            A_k*x <= b_k +1

        non-redundant if optimal solution > b_k
        """
        variables = list(polytope.variables)
        obj = []
        A = []
        b = []
        b_k = None

        for i, bound in enumerate(polytope.bounds):
            if i == index_to_check or i in non_redundant_index:
                a = [0] * len(variables)
                for var_name in bound.coefficients:
                    var_index = variables.index(var_name)
                    a[var_index] = bound.coefficients[var_name]
                A.append(a)
                b_i = bound.constant
                if i == index_to_check:
                    b_k = b_i
                    b_i += 1
                    obj = [1 * v for v in a]
                b.append(b_i)
        assert b_k is not None

        optimal_value, optimal_solution = self._lp(A, b, obj, "maximize")

        non_redundant = optimal_value > b_k

        if non_redundant:
            return True, self._ray_shoot(
                polytope, internal_point, optimal_solution, index_to_check
            )
        else:
            return False, index_to_check

    def _get_truth_values(self, polytope, point):
        values = []
        variables = list(polytope.variables)

        assert len(point) == len(variables)

        for index, bound in enumerate(polytope.bounds):
            coefficients = [0] * len(variables)
            for var_name in bound.coefficients:
                var_index = variables.index(var_name)
                coefficients[var_index] = bound.coefficients[var_name]
            polynomial = [point[i] * coefficients[i] for i in range(len(point))]
            truth_value = sum(polynomial) <= bound.constant
            values.append(truth_value)
        return values

    def _ray_shoot(self, polytope, start_point, end_point, index):
        values = self._get_truth_values(polytope, end_point)
        others = values[:index] + values[index + 1:]

        # if at the end point (optimal) there is only one inequality falsified
        # then return that particular inequality (index)
        if min(others) == max(others):
            return index
        try:
            return self._ray_shoot_iter(polytope, start_point, end_point)
        except RecursionError:
            print("RECURSION")
            return index

    def _ray_shoot_iter(self, polytope, start_point, end_point):
        # start point is inside the polytope so every bound is respected
        # calculate middle point
        assert len(start_point) == len(end_point)
        middle_point = [
            ((end_point[i] + start_point[i]) * Fraction(1, 2))
            for i in range(len(start_point))
        ]

        # check bounds
        intersected = None
        values = self._get_truth_values(polytope, middle_point)

        for i, v in enumerate(values):
            if not v:
                if intersected is None:
                    intersected = i
                else:
                    intersected = -1
                    break

        if intersected is None:
            return self._ray_shoot_iter(polytope, middle_point, end_point)
        elif intersected < 0:
            return self._ray_shoot_iter(polytope, start_point, middle_point)
        else:
            return intersected

    def _lp(self, A, B, obj, type_="maximize"):
        f = NamedTemporaryFile(mode="w+t", dir=("."))

        assert len(A) == len(B)

        variable_names = [("x_{}".format(i)) for i in range(len(A[0]))]

        # needed to retrieve the values
        f.writelines("(set-option :produce-models true)")

        # declare all variables
        for var in variable_names:
            f.writelines("(declare-fun {} () Real)".format(var))

        # add all constraints
        f.writelines("(assert (and")
        for i in range(len(A)):
            a = A[i]
            b = B[i]
            assert len(a) == len(variable_names)
            coeffs = []
            for c in a:
                if c < 0:
                    coeffs += ["(- {})".format(abs(c))]
                else:
                    coeffs += [str(c)]
            assert len(a) == len(coeffs)
            b = str(b) if b >= 0 else "(- {})".format(abs(b))
            monomials = [
                ("(* {} {})".format(coeffs[j], variable_names[j]))
                for j in range(len(a))
            ]
            if len(monomials) > 1:
                constraint = "(<= (+ {}) {})".format(" ".join(monomials), b)
            else:
                constraint = "(<= {} {})".format(monomials[0], b)
            f.writelines(constraint)
        f.writelines("))")

        # add objective
        coeffs = []
        for c in obj:
            if c < 0:
                coeffs += ["(- {})".format(abs(c))]
            else:
                coeffs += [str(c)]
        monomials = [
            ("(* {} {})".format(coeffs[j], variable_names[j])) for j in range(len(obj))
        ]
        if len(monomials) > 1:
            f.writelines("({} (+ {}))".format(type_, " ".join(monomials)))
        else:
            f.writelines("({} {})".format(type_, monomials[0]))

        f.writelines("(check-sat)")
        f.writelines("(get-objectives)")
        f.writelines("(load-objective-model 1)")
        for var in variable_names:
            f.writelines("(get-value ({}))".format(var))
        f.writelines("(exit)")
        f.seek(0)

        # read output
        values = []
        out = NamedTemporaryFile(mode="w+t", dir=("."))
        output = call(["optimathsat", f.name], stdout=out)
        out.seek(0)
        output = out.read()
        for line in output.split("\n"):
            """
            output can have these forms:
                ( (x_N DIGITS) )
                ( (x_N (- DIGITS)) )
                ( (x_N (/ DIGITS DIGITS)) )
                ( (x_N (- (/ DIGITS DIGITS))) )

            regex below is x_N + OR of the four different types
            """
            r = re.search(
                r"\( \(x_(\d+) (?:(\d+)|(?:\(- (\d+)\))|(?:\(\/ (\d+) (\d+)\))|(?:\(- \(\/ (\d+) (\d+)\)\)))\) \)",
                line,
            )
            if r:
                var_index = r.group(1)
                if r.group(2):
                    values.append(Fraction(int(r.group(2))))
                elif r.group(3):
                    values.append(-1 * Fraction(int(r.group(3))))
                elif r.group(4):
                    values.append(Fraction(int(r.group(4)), int(r.group(5))))
                elif r.group(6):
                    values.append(-1 * Fraction(int(r.group(6)), int(r.group(7))))

        assert len(values) == len(variable_names)
        obj_value = sum([(obj[i] * values[i]) for i in range(len(values))])
        return obj_value, values


class HashTable:
    def __init__(self):
        manager = Manager()
        self.table = manager.dict()

    def get(self, key):
        try:
            return self.table.get(key)
        except TypeError as ex:
            print("HashTable error:\n", ex)
            return self.get(key)

    def set(self, key, value):
        self.table[key] = value

    def key(self, polytope, cond_assignment, variables):
        bounds_key = []
        for index, bound in enumerate(polytope.bounds):
            bound_key = []
            for var in variables:
                if var in bound.coefficients:
                    bound_key.append(bound.coefficients[var])
                else:
                    bound_key.append(0)
            bound_key.append(bound.constant)
            bounds_key.append(tuple(bound_key))

        bounds_key = tuple(sorted(bounds_key))

        key = tuple([bounds_key, cond_assignment])
        return key

    def key_2(self, polytope, integrand):
        variables = list(integrand.variables.union(polytope.variables))

        # polytope key
        polytope_key = []
        for index, bound in enumerate(polytope.bounds):
            bound_key = []
            for var in variables:
                if var in bound.coefficients:
                    bound_key.append(bound.coefficients[var])
                else:
                    bound_key.append(0)
            bound_key.append(bound.constant)
            polytope_key.append(tuple(bound_key))

        polytope_key = tuple(sorted(polytope_key))

        # integrand key
        integrand_key = []

        if not isinstance(integrand, Polynomial):
            return integrand
        monomials = integrand.monomials

        for mon in monomials:
            mon_key = [float(mon.coefficient)]
            for var in variables:
                if var in mon.exponents:
                    mon_key.append(float(mon.exponents[var]))
                else:
                    mon_key.append(0)
            integrand_key.append(tuple(mon_key))

        integrand_key = tuple(sorted(integrand_key))

        key = tuple([polytope_key, integrand_key])
        return key



# from pysmt.shortcuts import Bool
import sys

from pywmi.domain import Domain
from pywmi.engines import PyXaddEngine

from wmipa.integration.cache_integrator import CacheIntegrator
from wmipa.integration.expression import Expression
from wmipa.integration.polytope import Polynomial, Polytope

sys.setrecursionlimit(10 ** 5)


class SymbolicIntegrator(CacheIntegrator):

    @staticmethod
    def _integrator():
        return PyXaddEngine
        # return partial(
        #     XsddEngine,
        #     algebra=PyXaddAlgebra(symbolic_backend=SympyAlgebra()),
        #     ordered=False,
        # )

        # return partial(
        #     FXSDD,
        #     vtree_strategy=balanced,
        #     algebra=PyXaddAlgebra(symbolic_backend=SympyAlgebra()),
        #     ordered=False,
        # )
        # return XaddEngine

    @classmethod
    def _make_problem(cls, weight, bounds, aliases):
        """Makes the problem to be solved by a symbolic solver.

        Args:
            weight (FNode): The weight function.
            bounds (list): The polytope.
            aliases (dict): The aliases of the variables.
        Returns:
            integrand (Expression): The integrand.
            polytope (Polytope): The polytope.
        """
        integrand = Expression(weight, aliases)
        polytope = Polytope(bounds, aliases)

        return integrand, polytope

    def _integrate_problem(self, integrand: Polynomial, polytope: Polytope):
        """Computes the integral of `integrand` over `polytope` using symbolic
            integration methods

        Args:
            integrand (Polynomial): The integrand of the integration.
            polytope (Polytope): The polytope of the integration.

        Returns:
            real: The integration result.

        """
        support = polytope.to_pysmt()
        weight = integrand.to_pysmt()
        variables = {
            v: (None, None) for v in integrand.variables.union(polytope.variables)
        }
        domain = Domain.make(real_variables=variables)
        return SymbolicIntegrator._integrator()(
            domain=domain, support=support, weight=weight
        ).compute_volume(add_bounds=False)
